from collections.abc import AsyncGenerator, AsyncIterator, Sequence
from datetime import UTC, datetime
from typing import Any, Literal, overload

from haiway import AsyncQueue, ctx

from draive.conversation.types import ConversationMemory, ConversationMessage
from draive.instructions import Instruction
from draive.lmm import (
    LMMCompletion,
    LMMContextElement,
    LMMInput,
    LMMStreamChunk,
    LMMStreamInput,
    LMMStreamProperties,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponse,
    LMMToolResponses,
    lmm_invoke,
    lmm_stream,
)
from draive.multimodal import Multimodal, MultimodalContent
from draive.prompts import Prompt
from draive.tools import Toolbox
from draive.utils import Memory

__all__ = [
    "default_conversation_completion",
]


@overload
async def default_conversation_completion(
    *,
    instruction: Instruction | str | None,
    input: ConversationMessage | Prompt | Multimodal,
    memory: ConversationMemory,
    toolbox: Toolbox,
    stream: Literal[True],
    **extra: Any,
) -> AsyncIterator[LMMStreamChunk]: ...


@overload
async def default_conversation_completion(
    *,
    instruction: Instruction | str | None,
    input: ConversationMessage | Prompt | Multimodal,
    memory: ConversationMemory,
    toolbox: Toolbox,
    stream: Literal[False] = False,
    **extra: Any,
) -> ConversationMessage: ...


@overload
async def default_conversation_completion(
    *,
    instruction: Instruction | str | None,
    input: ConversationMessage | Prompt | Multimodal,
    memory: ConversationMemory,
    toolbox: Toolbox,
    stream: bool,
    **extra: Any,
) -> AsyncIterator[LMMStreamChunk] | ConversationMessage: ...


async def default_conversation_completion(
    *,
    instruction: Instruction | str | None,
    input: ConversationMessage | Prompt | Multimodal,  # noqa: A002
    memory: ConversationMemory,
    toolbox: Toolbox,
    stream: bool = False,
    **extra: Any,
) -> AsyncIterator[LMMStreamChunk] | ConversationMessage:
    with ctx.scope("conversation_completion"):
        recalled_messages: Sequence[ConversationMessage] = await memory.recall()
        context: list[LMMContextElement]
        match input:
            case ConversationMessage() as message:
                await memory.remember(message)
                context = [
                    *(message.as_lmm_context_element() for message in recalled_messages),
                    message.as_lmm_context_element(),
                ]

            case Prompt() as prompt:
                prompt_messages: list[ConversationMessage] = []
                for element in prompt.content:
                    match element:
                        case LMMCompletion() as completion_element:
                            prompt_messages.append(
                                ConversationMessage(
                                    role="model",
                                    created=datetime.now(UTC),
                                    content=completion_element.content,
                                )
                            )

                        case LMMInput() as input_element:
                            prompt_messages.append(
                                ConversationMessage(
                                    role="user",
                                    created=datetime.now(UTC),
                                    content=input_element.content,
                                )
                            )

                        case _:
                            pass  # TODO add other content to memory when able

                await memory.remember(*prompt_messages)

                context = [
                    *(message.as_lmm_context_element() for message in recalled_messages),
                    *prompt.content,
                ]

            case content:
                message = ConversationMessage(
                    role="user",
                    created=datetime.now(UTC),
                    content=MultimodalContent.of(content),
                )
                await memory.remember(message)
                context = [
                    *(message.as_lmm_context_element() for message in recalled_messages),
                    message.as_lmm_context_element(),
                ]

        if stream:
            return ctx.stream(
                _conversation_stream,
                instruction=instruction,
                memory=memory,
                context=context,
                toolbox=toolbox,
                **extra,
            )

        else:
            return await _conversation_completion(
                instruction=instruction,
                memory=memory,
                context=context,
                toolbox=toolbox,
                **extra,
            )


async def _conversation_completion(
    instruction: Instruction | str | None,
    memory: Memory[Sequence[ConversationMessage], ConversationMessage],
    context: list[LMMContextElement],
    toolbox: Toolbox,
    **extra: Any,
) -> ConversationMessage:
    recursion_level: int = 0
    while recursion_level <= toolbox.repeated_calls_limit:
        match await lmm_invoke(
            instruction=instruction,
            context=context,
            tools=toolbox.available_tools(),
            tool_selection=toolbox.tool_selection(repetition_level=recursion_level),
            output="text",
            **extra,
        ):
            case LMMCompletion() as completion:
                ctx.log_debug("Received conversation result")
                response_message: ConversationMessage = ConversationMessage(
                    role="model",
                    created=datetime.now(UTC),
                    content=completion.content,
                )
                await memory.remember(response_message)
                return response_message

            case LMMToolRequests() as tool_requests:
                ctx.log_debug("Received conversation tool calls")

                responses: Sequence[LMMToolResponse] = await toolbox.respond_all(tool_requests)

                if direct_content := [
                    response.content for response in responses if response.direct
                ]:
                    response_message: ConversationMessage = ConversationMessage(
                        role="model",
                        created=datetime.now(UTC),
                        content=MultimodalContent.of(*direct_content),
                    )
                    await memory.remember(response_message)
                    return response_message

                else:
                    context.extend(
                        [
                            tool_requests,
                            LMMToolResponses(responses=responses),
                        ]
                    )

        recursion_level += 1  # continue with next recursion level

    raise RuntimeError("LMM exceeded limit of recursive calls")


async def _conversation_stream(
    instruction: Instruction | str | None,
    message: ConversationMessage,
    memory: Memory[Sequence[ConversationMessage], ConversationMessage],
    context: list[LMMContextElement],
    toolbox: Toolbox,
    **extra: Any,
) -> AsyncGenerator[LMMStreamChunk]:
    input_stream = AsyncQueue[LMMStreamInput]()
    input_stream.enqueue(
        LMMStreamChunk.of(
            message.content,
            eod=True,  # we provide single input chunk here
        )
    )
    accumulated_content: MultimodalContent = MultimodalContent.of()
    async for element in await lmm_stream(
        properties=LMMStreamProperties(
            instruction=instruction,
            tools=toolbox.available_tools(),
        ),
        input=input_stream,
        context=context,
        **extra,
    ):
        match element:
            case LMMStreamChunk() as chunk:
                accumulated_content = accumulated_content.appending(
                    chunk.content,
                    merge_text=True,
                )

                yield chunk

                # on turn end finalize the stream - we are streaming only a single response here
                if chunk.eod:
                    input_stream.finish()
                    response_message: ConversationMessage = ConversationMessage(
                        role="model",
                        created=datetime.now(UTC),
                        content=accumulated_content,
                    )
                    await memory.remember(
                        message,
                        response_message,
                    )

                    return  # end of streaming for conversation completion

            case LMMToolRequest() as tool_request:
                ctx.spawn(
                    handle_tool_call,
                    tool_request,
                    toolbox=toolbox,
                    output=input_stream,
                )


async def handle_tool_call(
    request: LMMToolRequest,
    /,
    *,
    toolbox: Toolbox,
    output: AsyncQueue[LMMStreamInput],
) -> None:
    output.enqueue(await toolbox.respond(request))



from collections.abc import Sequence
from typing import Any, Protocol, Self, runtime_checkable

from haiway import State

from draive.multimodal import (
    MediaContent,
    MultimodalContentElement,
    TextContent,
    validated_media_type,
)
from draive.parameters import DataModel

__all__ = [
    "MissingResource",
    "Resource",
    "ResourceContent",
    "ResourceDeclaration",
    "ResourceFetching",
    "ResourceListing",
]


class MissingResource(Exception):
    pass


class ResourceDeclaration(DataModel):
    uri: str
    mime_type: str | None
    name: str
    description: str | None = None


# TODO: add ResourceTemplateDeclaration for ResourceTemplates support
# https://modelcontextprotocol.io/docs/concepts/resources#resource-templates


class ResourceContent(State):
    mime_type: str
    blob: bytes

    def as_multimodal(self) -> MultimodalContentElement:
        match self.mime_type:
            case "text/plain":
                return TextContent(text=self.blob.decode())

            case "application/json":
                return DataModel.from_json(self.blob.decode())

            case other:
                # try to match supported media or raise an exception
                return MediaContent.data(
                    self.blob,
                    media=validated_media_type(other),
                )


class Resource(State):
    uri: str
    name: str
    description: str | None
    content: Sequence[Self] | ResourceContent


@runtime_checkable
class ResourceListing(Protocol):
    async def __call__(
        self,
        **extra: Any,
    ) -> Sequence[ResourceDeclaration]: ...


@runtime_checkable
class ResourceFetching(Protocol):
    async def __call__(
        self,
        uri: str,
        **extra: Any,
    ) -> Resource | None: ...



from draive.helpers.volatile_index import (
    VectorIndex,
    VectorIndexing,
    VectorSearching,
)
from draive.helpers.volatile_memory import VolatileMemory

__all__ = [
    "VectorIndex",
    "VectorIndexing",
    "VectorSearching",
    "VolatileMemory",
]



from collections.abc import AsyncIterator, Mapping, Sequence
from datetime import datetime
from typing import Any, Literal, Protocol, Self, overload, runtime_checkable
from uuid import uuid4

from draive.instructions import Instruction
from draive.lmm import (
    LMMCompletion,
    LMMContextElement,
    LMMInput,
    LMMStreamChunk,
)
from draive.multimodal import (
    Multimodal,
    MultimodalContent,
)
from draive.parameters import DataModel, Field
from draive.prompts import Prompt
from draive.tools import Toolbox
from draive.utils import Memory

__all__ = [
    "ConversationCompletion",
    "ConversationElement",
    "ConversationMemory",
    "ConversationMessage",
]


class ConversationMessage(DataModel):
    @classmethod
    def user(
        cls,
        content: Multimodal,
        identifier: str | None = None,
        author: str | None = None,
        created: datetime | None = None,
        meta: Mapping[str, str | float | int | bool | None] | None = None,
    ) -> Self:
        return cls(
            identifier=identifier or uuid4().hex,
            role="user",
            author=author,
            created=created,
            content=MultimodalContent.of(content),
            meta=meta,
        )

    @classmethod
    def model(
        cls,
        content: Multimodal,
        identifier: str | None = None,
        author: str | None = None,
        created: datetime | None = None,
        meta: Mapping[str, str | float | int | bool | None] | None = None,
    ) -> Self:
        return cls(
            identifier=identifier or uuid4().hex,
            role="model",
            author=author,
            created=created,
            content=MultimodalContent.of(content),
            meta=meta,
        )

    identifier: str = Field(default_factory=lambda: uuid4().hex)
    role: Literal["user", "model"]
    author: str | None = None
    created: datetime | None = None
    content: MultimodalContent
    meta: Mapping[str, str | float | int | bool | None] | None = None

    def as_lmm_context_element(self) -> LMMContextElement:
        match self.role:
            case "user":
                return LMMInput.of(self.content)

            case "model":
                return LMMCompletion.of(self.content)

    def __bool__(self) -> bool:
        return bool(self.content)


ConversationElement = ConversationMessage  # TODO: allow events/tool statuses
ConversationMemory = Memory[Sequence[ConversationElement], ConversationElement]


@runtime_checkable
class ConversationCompletion(Protocol):
    @overload
    async def __call__(
        self,
        *,
        instruction: Instruction | str | None,
        input: ConversationMessage | Prompt | Multimodal,
        memory: Memory[Sequence[ConversationMessage], ConversationMessage],
        toolbox: Toolbox,
        stream: Literal[True],
        **extra: Any,
    ) -> AsyncIterator[LMMStreamChunk]: ...

    @overload
    async def __call__(
        self,
        *,
        instruction: Instruction | str | None,
        input: ConversationMessage | Prompt | Multimodal,
        memory: Memory[Sequence[ConversationMessage], ConversationMessage],
        toolbox: Toolbox,
        stream: Literal[False] = False,
        **extra: Any,
    ) -> ConversationMessage: ...

    @overload
    async def __call__(
        self,
        *,
        instruction: Instruction | str | None,
        input: ConversationMessage | Prompt | Multimodal,
        memory: Memory[Sequence[ConversationMessage], ConversationMessage],
        toolbox: Toolbox,
        stream: bool,
        **extra: Any,
    ) -> AsyncIterator[LMMStreamChunk] | ConversationMessage: ...

    async def __call__(
        self,
        *,
        instruction: Instruction | str | None,
        input: ConversationMessage | Prompt | Multimodal,  # noqa: A002
        memory: Memory[Sequence[ConversationMessage], ConversationMessage],
        toolbox: Toolbox,
        stream: bool = False,
        **extra: Any,
    ) -> AsyncIterator[LMMStreamChunk] | ConversationMessage: ...



from draive.bedrock.client import BedrockClient
from draive.bedrock.config import BedrockChatConfig
from draive.bedrock.lmm import bedrock_lmm
from draive.bedrock.types import BedrockException

__all__ = [
    "BedrockChatConfig",
    "BedrockClient",
    "BedrockException",
    "bedrock_lmm",
]



from collections.abc import Sequence
from typing import Any, Protocol, Self, final, runtime_checkable

from haiway import State

__all__ = [
    "Memory",
]


@runtime_checkable
class MemoryRecall[Recalled](Protocol):
    async def __call__(
        self,
        **extra: Any,
    ) -> Recalled: ...


@runtime_checkable
class MemoryRemember[Remembered](Protocol):
    async def __call__(
        self,
        *items: Remembered,
        **extra: Any,
    ) -> None: ...


@final
class Memory[Recalled, Remembered](State):
    @classmethod
    def constant(
        cls,
        recalled: Recalled,
    ) -> Self:
        async def recall(
            **extra: Any,
        ) -> Recalled:
            return recalled

        async def remember(
            *items: Remembered,
            **extra: Any,
        ) -> None:
            pass  # noop

        return cls(
            recall=recall,
            remember=remember,
        )

    @classmethod
    def volatile[Item](
        cls,
        /,
        *,
        initial: Item,
        limit: int | None = None,
    ) -> "Memory[Item, Item]":
        storage: Item = initial
        limit = limit or 0

        async def recall(
            **extra: Any,
        ) -> Item:
            nonlocal storage
            return storage

        async def remember(
            *items: Item,
            **extra: Any,
        ) -> None:
            nonlocal storage
            if not items:
                return  # nothing to do

            storage = items[-1]

        return Memory[Item, Item](
            recall=recall,
            remember=remember,
        )

    @classmethod
    def accumulative_volatile[Item](
        cls,
        /,
        *,
        initial: Sequence[Item] | None = None,
        limit: int | None = None,
    ) -> "Memory[Sequence[Item], Item]":
        storage: list[Item] = list(initial) if initial else []
        limit = limit or 0

        async def recall(
            **extra: Any,
        ) -> Sequence[Item]:
            nonlocal storage
            return storage

        async def remember(
            *items: Item,
            **extra: Any,
        ) -> None:
            nonlocal storage
            if not items:
                return  # nothing to do

            storage.extend(items)

            if limit > 0:
                storage = storage[-limit:]

        return Memory[Sequence[Item], Item](
            recall=recall,
            remember=remember,
        )

    recall: MemoryRecall[Recalled]
    remember: MemoryRemember[Remembered]



from base64 import b64encode
from collections.abc import Iterable, Sequence
from copy import copy
from typing import Any, cast, get_args
from uuid import uuid4

from haiway import ArgumentsTrace, ResultTrace, ctx

from draive.gemini.client import GeminiClient
from draive.gemini.config import GeminiConfig
from draive.gemini.models import (
    GeminiChoice,
    GeminiDataMessageContent,
    GeminiDataReferenceMessageContent,
    GeminiFunctionCallMessageContent,
    GeminiFunctionsTool,
    GeminiFunctionToolSpecification,
    GeminiGenerationResult,
    GeminiMessage,
    GeminiMessageContent,
    GeminiRequestMessage,
    GeminiTextMessageContent,
)
from draive.gemini.types import GeminiException
from draive.instructions import Instruction
from draive.lmm import (
    LMMCompletion,
    LMMContext,
    LMMContextElement,
    LMMInput,
    LMMInvocation,
    LMMOutput,
    LMMOutputSelection,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponses,
    LMMToolSelection,
    LMMToolSpecification,
)
from draive.metrics import TokenUsage
from draive.multimodal import MediaContent, MultimodalContent, MultimodalContentElement, TextContent
from draive.multimodal.media import MediaType
from draive.parameters import DataModel

__all__ = [
    "gemini_lmm",
]


def gemini_lmm(
    client: GeminiClient | None = None,
    /,
) -> LMMInvocation:
    client = client or GeminiClient.shared()

    async def lmm_invocation(
        *,
        instruction: Instruction | str | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        **extra: Any,
    ) -> LMMOutput:
        with ctx.scope("gemini_lmm_invocation"):
            ctx.record(
                ArgumentsTrace.of(
                    instruction=instruction,
                    context=context,
                    tool_selection=tool_selection,
                    tools=tools,
                    output=output,
                    **extra,
                ),
            )

            config: GeminiConfig = ctx.state(GeminiConfig).updated(**extra)
            ctx.record(config)

            match output:
                case "auto" | "text":
                    config = config.updated(response_format="text/plain")

                case "image":
                    raise NotImplementedError("image output is not supported by gemini")

                case "audio":
                    raise NotImplementedError("audio output is not supported by gemini")

                case "video":
                    raise NotImplementedError("video output is not supported by gemini")

                case _:  # TODO: utilize json schema within API
                    if tools:
                        ctx.log_warning(
                            "Attempting to use Gemini in JSON mode with tools which is"
                            " not supported. Using text mode instead..."
                        )
                        config = config.updated(response_format="text/plain")

                    else:
                        config = config.updated(response_format="application/json")

            return await _generate(
                client=client,
                config=config,
                instruction=Instruction.formatted(instruction) or "",
                messages=[_convert_context_element(element=element) for element in context],
                tool_selection=tool_selection,
                tools=tools,
            )

    return LMMInvocation(invoke=lmm_invocation)


def _convert_content_element(
    element: MultimodalContentElement,
) -> dict[str, Any]:
    match element:
        case TextContent() as text:
            return {"text": text.text}

        case MediaContent() as media:
            match media.source:
                case str() as reference:
                    return {
                        "fileData": {
                            "mimeType": media.media,
                            "fileUri": reference,
                        }
                    }

                case bytes() as data:
                    return {
                        "inlineData": {
                            "mimeType": media.media,
                            "data": b64encode(data).decode(),
                        }
                    }

        case DataModel() as data:
            return {"text": data.as_json()}


def _convert_context_element(
    element: LMMContextElement,
) -> GeminiRequestMessage:
    match element:
        case LMMInput() as input:
            return {
                "role": "user",
                "parts": [
                    _convert_content_element(element=element) for element in input.content.parts
                ],
            }

        case LMMCompletion() as completion:
            return {
                "role": "model",
                "parts": [
                    _convert_content_element(element=element)
                    for element in completion.content.parts
                ],
            }

        case LMMToolRequests() as tool_requests:
            return {
                "role": "model",
                "parts": [
                    {
                        "functionCall": {
                            "name": request.tool,
                            "args": request.arguments,
                        },
                    }
                    for request in tool_requests.requests
                ],
            }

        case LMMToolResponses() as tool_responses:
            return {
                "role": "model",
                "parts": [
                    {
                        "functionResponse": {
                            "name": response.tool,
                            "response": response.content.as_dict(),
                        },
                    }
                    for response in tool_responses.responses
                ],
            }


def _convert_content_part(
    part: GeminiMessageContent,
) -> MultimodalContentElement:
    match part:
        case GeminiTextMessageContent() as text:
            return TextContent(text=text.text)

        case GeminiDataMessageContent() as data:
            mime_type: str = data.data.mime_type
            if mime_type in get_args(MediaType):
                return MediaContent.base64(
                    data.data.data,
                    media=cast(MediaType, mime_type),
                )

            else:
                raise GeminiException("Unsupported result content data %s", data)

        case GeminiDataReferenceMessageContent() as reference:
            mime_type: str = reference.reference.mime_type
            if mime_type in get_args(MediaType):
                return MediaContent.url(
                    reference.reference.uri,
                    media=cast(MediaType, mime_type),
                )

            else:
                raise GeminiException("Unsupported result content reference %s", reference)

        case other:
            raise GeminiException("Unsupported result content %s", other)


async def _generate(  # noqa: PLR0913, C901, PLR0912, PLR0915
    *,
    client: GeminiClient,
    config: GeminiConfig,
    instruction: str,
    messages: list[GeminiRequestMessage],
    tool_selection: LMMToolSelection,
    tools: Iterable[LMMToolSpecification] | None,
) -> LMMOutput:
    result: GeminiGenerationResult
    converted_tools: Sequence[GeminiFunctionToolSpecification] = []
    for tool in tools or []:
        tool_function: GeminiFunctionToolSpecification = {
            "name": tool["name"],
            "description": tool["description"] or "",
            "parameters": cast(dict[str, Any], tool["parameters"]),
        }
        # AIStudio api requires to delete properties if those are empty...
        if "parameters" in tool_function and not tool_function["parameters"]["properties"]:
            tool_function = copy(tool_function)
            del tool_function["parameters"]

        converted_tools.append(tool_function)

    match tool_selection:
        case "auto":
            result = await client.generate(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[GeminiFunctionsTool(functionDeclarations=converted_tools)],
                tool_calling_mode="AUTO",
            )

        case "none":
            result = await client.generate(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[],
                tool_calling_mode="NONE",
            )

        case "required":
            result = await client.generate(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[GeminiFunctionsTool(functionDeclarations=converted_tools)],
                tool_calling_mode="ANY",
            )

        case tool:
            assert tool in (tools or []), "Can't suggest a tool without using it"  # nosec: B101
            tool_function: GeminiFunctionToolSpecification = {
                "name": tool["name"],
                "description": tool["description"] or "",
                "parameters": cast(dict[str, Any], tool["parameters"]),
            }

            # AIStudio api requires to delete properties if those are empty...
            if "parameters" in tool_function and not tool_function["parameters"]["properties"]:
                tool_function = copy(tool_function)
                del tool_function["parameters"]

            result = await client.generate(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[GeminiFunctionsTool(functionDeclarations=[tool_function])],
                tool_calling_mode="ANY",
            )

    if usage := result.usage:
        ctx.record(
            TokenUsage.for_model(
                config.model,
                input_tokens=usage.prompt_tokens,
                output_tokens=usage.generated_tokens,
            ),
        )

    if not result.choices:
        raise GeminiException("Invalid Gemini completion - missing messages!", result)

    result_choice: GeminiChoice = result.choices[0]
    result_message: GeminiMessage
    match result_choice.finish_reason:
        case "STOP":
            if message := result_choice.content:
                result_message = message

            else:
                raise GeminiException(
                    "Invalid Gemini response - missing message content: %s", result
                )

        case "MAX_TOKENS":
            raise GeminiException("Gemini response finish caused by token limit: %s", result)

        case "SAFETY":
            raise GeminiException("Gemini response finish caused by safety reason: %s", result)

        case "RECITATION":
            raise GeminiException("Gemini response finish caused by recitation reason: %s", result)

        case "OTHER":
            raise GeminiException("Gemini response finish caused by unknown reason: %s", result)

    message_parts: list[
        GeminiTextMessageContent | GeminiDataReferenceMessageContent | GeminiDataMessageContent
    ] = []

    tool_calls: list[GeminiFunctionCallMessageContent] = []
    for part in result_message.content:
        match part:
            case GeminiTextMessageContent() as text:
                message_parts.append(text)

            case GeminiFunctionCallMessageContent() as call:
                tool_calls.append(call)

            case GeminiDataReferenceMessageContent() as reference:
                message_parts.append(reference)

            case GeminiDataMessageContent() as data:
                message_parts.append(data)

            case other:
                raise GeminiException("Invalid Gemini completion part", other)

    if tool_calls and (tools := tools):
        if __debug__ and message_parts:
            ctx.log_debug("Gemini has generated a message and tool calls, ignoring the message...")

        ctx.record(ResultTrace.of(tool_calls))

        return LMMToolRequests(
            requests=[
                LMMToolRequest(
                    identifier=uuid4().hex,
                    tool=call.function_call.name,
                    arguments=call.function_call.arguments,
                )
                for call in tool_calls
            ]
        )

    elif message_parts:
        ctx.record(ResultTrace.of(message_parts))
        return LMMCompletion.of(
            MultimodalContent.of(
                *[_convert_content_part(part) for part in message_parts],
            )
        )

    else:
        raise GeminiException("Invalid Gemini completion", result)



from haiway import State

from draive.generation.text.default import default_generate_text
from draive.generation.text.types import TextGenerator

__all__ = [
    "TextGeneration",
]


class TextGeneration(State):
    generate: TextGenerator = default_generate_text



from collections.abc import Sequence
from typing import Any

from haiway import ctx

from draive.embedding import Embedded, TextEmbedding
from draive.gemini.client import GeminiClient
from draive.gemini.config import GeminiEmbeddingConfig

__all__ = [
    "gemini_text_embedding",
]


def gemini_text_embedding(
    client: GeminiClient | None = None,
    /,
) -> TextEmbedding:
    client = client or GeminiClient.shared()

    async def gemini_embed_text(
        values: Sequence[str],
        **extra: Any,
    ) -> Sequence[Embedded[str]]:
        config: GeminiEmbeddingConfig = ctx.state(GeminiEmbeddingConfig).updated(**extra)
        with ctx.scope("embed_text"):
            results: list[list[float]] = await client.embedding(
                config=config,
                inputs=values,
            )

            return [
                Embedded(
                    value=embedded[0],
                    vector=embedded[1],
                )
                for embedded in zip(
                    values,
                    results,
                    strict=True,
                )
            ]

    return TextEmbedding(embed=gemini_embed_text)



from draive.openai.client import OpenAIClient
from draive.openai.config import (
    OpenAIChatConfig,
    OpenAIEmbeddingConfig,
    OpenAIImageGenerationConfig,
    OpenAIModerationConfig,
)
from draive.openai.embedding import openai_text_embedding
from draive.openai.guardrails import openai_content_guardrails
from draive.openai.images import openai_image_generator
from draive.openai.lmm import openai_lmm, openai_streaming_lmm
from draive.openai.tokenization import openai_tokenizer
from draive.openai.types import OpenAIException

__all__ = [
    "OpenAIChatConfig",
    "OpenAIClient",
    "OpenAIEmbeddingConfig",
    "OpenAIException",
    "OpenAIImageGenerationConfig",
    "OpenAIModerationConfig",
    "openai_content_guardrails",
    "openai_image_generator",
    "openai_lmm",
    "openai_streaming_lmm",
    "openai_text_embedding",
    "openai_tokenizer",
]



from asyncio import gather
from collections.abc import Callable, Mapping
from typing import Protocol, Self, cast, final, overload, runtime_checkable

from haiway import AttributePath, ctx, freeze

from draive.evaluation.score import EvaluationScore
from draive.evaluation.value import EvaluationScoreValue, evaluation_score_value
from draive.parameters import DataModel, Field

__all__ = [
    "Evaluator",
    "EvaluatorDefinition",
    "EvaluatorResult",
    "PreparedEvaluator",
    "evaluator",
    "evaluator_highest",
    "evaluator_lowest",
]


def _verifier(
    value: float,
) -> None:
    if not (0 <= value <= 1):
        raise ValueError(f"Threshold has to have a value between 0 and 1, received: {value}")


class EvaluatorResult(DataModel):
    @classmethod
    def of(
        cls,
        evaluator: str,
        /,
        *,
        score: EvaluationScoreValue,
        score_comment: str | None = None,
        threshold: EvaluationScoreValue,
        meta: Mapping[str, str | float | int | bool | None] | None = None,
    ) -> Self:
        return cls(
            evaluator=evaluator,
            score=EvaluationScore.of(
                score,
                comment=score_comment,
            ),
            threshold=evaluation_score_value(threshold),
            meta=meta,
        )

    evaluator: str = Field(
        description="Name of evaluator",
    )
    score: EvaluationScore = Field(
        description="Evaluation score",
    )
    threshold: float = Field(
        description="Score threshold required to pass evaluation, "
        "a value between 0 (min) and 1 (max)",
        verifier=_verifier,
    )
    meta: Mapping[str, str | float | int | bool | None] | None = Field(
        description="Additional evaluation metadata",
        default=None,
    )

    @property
    def passed(self) -> bool:
        return self.score.value >= self.threshold

    def report(self) -> str:
        meta_values: str = (
            f"\n{'\n'.join(f'{key}: {value}' for key, value in self.meta.items())}"
            if self.meta
            else "N/A"
        )
        return (
            f"{self.evaluator} {'passed' if self.passed else 'failed'}"
            f" with score {self.score.value},"
            f" required {self.threshold},"
            f" comment: {f"'{self.score.comment}'" or 'N/A'}"
            f" meta:\n{meta_values}"
        )


class EvaluationResult(DataModel):
    @classmethod
    async def of(
        cls,
        score: EvaluationScore | float | bool,
        /,
        meta: Mapping[str, str | float | int | bool | None] | None = None,
    ) -> Self:
        evaluation_score: EvaluationScore
        match score:
            case EvaluationScore() as score:
                evaluation_score = score

            case float() as value:
                evaluation_score = EvaluationScore(value=value)

            case passed:
                evaluation_score = EvaluationScore(value=1.0 if passed else 0.0)

        return cls(
            score=evaluation_score,
            meta=meta,
        )

    score: EvaluationScore = Field(
        description="Evaluation score",
    )
    meta: Mapping[str, str | float | int | bool | None] | None = Field(
        description="Additional evaluation metadata",
        default=None,
    )


@runtime_checkable
class EvaluatorDefinition[Value, **Args](Protocol):
    @property
    def __name__(self) -> str: ...

    async def __call__(
        self,
        value: Value,
        /,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> EvaluationResult | EvaluationScore | float | bool: ...


@runtime_checkable
class PreparedEvaluator[Value](Protocol):
    async def __call__(
        self,
        value: Value,
        /,
    ) -> EvaluatorResult: ...


@final
class Evaluator[Value, **Args]:
    def __init__(
        self,
        name: str,
        definition: EvaluatorDefinition[Value, Args],
        threshold: float | None,
        meta: Mapping[str, str | float | int | bool | None] | None = None,
    ) -> None:
        assert (  # nosec: B101
            threshold is None or 0 <= threshold <= 1
        ), "Evaluation threshold has to be between 0 and 1"

        self._definition: EvaluatorDefinition[Value, Args] = definition
        self.name: str = name
        self.threshold: float = 1 if threshold is None else threshold
        self.meta: Mapping[str, str | float | int | bool | None] | None = meta

        freeze(self)

    def with_threshold(
        self,
        value: EvaluationScoreValue,
        /,
    ) -> Self:
        return self.__class__(
            name=self.name,
            definition=self._definition,
            threshold=evaluation_score_value(value),
            meta=self.meta,
        )

    def with_meta(
        self,
        meta: Mapping[str, str | float | int | bool | None],
        /,
    ) -> Self:
        return self.__class__(
            name=self.name,
            definition=self._definition,
            threshold=self.threshold,
            meta={**self.meta, **meta} if self.meta else meta,
        )

    def prepared(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> PreparedEvaluator[Value]:
        async def evaluate(
            value: Value,
        ) -> EvaluatorResult:
            return await self(
                value,
                *args,
                **kwargs,
            )

        return evaluate

    def contra_map[Mapped](
        self,
        mapping: Callable[[Mapped], Value] | AttributePath[Mapped, Value] | Value,
        /,
    ) -> "Evaluator[Mapped, Args]":
        mapper: Callable[[Mapped], Value]
        match mapping:
            case Callable() as function:  # pyright: ignore[reportUnknownVariableType]
                mapper = function

            case path:
                assert isinstance(  # nosec: B101
                    path, AttributePath
                ), "Prepare parameter path by using Self._.path.to.property"
                mapper = cast(AttributePath[Mapped, Value], path).__call__

        async def evaluation(
            value: Mapped,
            /,
            *args: Args.args,
            **kwargs: Args.kwargs,
        ) -> EvaluationResult | EvaluationScore | float | bool:
            return await self._definition(
                mapper(value),
                *args,
                **kwargs,
            )

        return Evaluator[Mapped, Args](
            name=self.name,
            definition=evaluation,
            threshold=self.threshold,
        )

    async def __call__(
        self,
        value: Value,
        /,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> EvaluatorResult:
        evaluation_score: EvaluationScore
        evaluation_meta: Mapping[str, str | float | int | bool | None] | None
        try:
            match await self._definition(
                value,
                *args,
                **kwargs,
            ):
                case EvaluationResult() as result:
                    evaluation_score = result.score
                    evaluation_meta = result.meta

                case EvaluationScore() as score:
                    evaluation_score = score
                    evaluation_meta = None

                case float() as score_value:
                    evaluation_score = EvaluationScore(value=score_value)
                    evaluation_meta = None

                case passed:
                    evaluation_score = EvaluationScore(value=1 if passed else 0)
                    evaluation_meta = None

        except Exception as exc:
            ctx.log_error(
                f"Evaluator `{self.name}` failed, using `0` score fallback result",
                exception=exc,
            )
            evaluation_score = EvaluationScore(
                value=0,
                comment="Evaluation failed",
            )
            evaluation_meta = {"exception": str(exc)}

        result_meta: Mapping[str, str | float | int | bool | None] | None
        if self.meta:
            if evaluation_meta:
                result_meta = {**self.meta, **evaluation_meta}

            else:
                result_meta = self.meta

        elif evaluation_meta:
            result_meta = evaluation_meta

        else:
            result_meta = None

        return EvaluatorResult(
            evaluator=self.name,
            score=evaluation_score,
            threshold=self.threshold,
            meta=result_meta,
        )


@overload
def evaluator[Value, **Args](
    definition: EvaluatorDefinition[Value, Args],
    /,
) -> Evaluator[Value, Args]: ...


@overload
def evaluator[Value, **Args](
    *,
    name: str | None = None,
    threshold: EvaluationScoreValue | None = None,
) -> Callable[
    [EvaluatorDefinition[Value, Args]],
    Evaluator[Value, Args],
]: ...


def evaluator[Value, **Args](  # pyright: ignore[reportInconsistentOverload] - this seems to be pyright false positive/error
    evaluation: EvaluatorDefinition[Value, Args] | None = None,
    /,
    *,
    name: str | None = None,
    threshold: EvaluationScoreValue | None = None,
) -> (
    Callable[
        [EvaluatorDefinition[Value, Args]],
        Evaluator[Value, Args],
    ]
    | Evaluator[Value, Args]
):
    def wrap(
        definition: EvaluatorDefinition[Value, Args],
    ) -> Evaluator[Value, Args]:
        return Evaluator(
            name=name or definition.__name__,
            definition=definition,
            threshold=evaluation_score_value(threshold) if threshold is not None else None,
        )

    if evaluation:
        return wrap(evaluation)

    else:
        return wrap


def evaluator_lowest[Value](
    evaluators: PreparedEvaluator[Value],
    /,
    *_evaluators: PreparedEvaluator[Value],
) -> PreparedEvaluator[Value]:
    async def evaluate(
        value: Value,
    ) -> EvaluatorResult:
        # Placeholder for the lowest result
        lowest: EvaluatorResult = EvaluatorResult(
            evaluator="lowest",
            score=EvaluationScore(value=1),
            threshold=0,
            meta=None,
        )

        for result in await gather(
            evaluators(value),
            *[evaluator(value) for evaluator in _evaluators],
        ):
            if result.score <= lowest.score:
                lowest = result

        return lowest

    return evaluate


def evaluator_highest[Value](
    evaluators: PreparedEvaluator[Value],
    /,
    *_evaluators: PreparedEvaluator[Value],
) -> PreparedEvaluator[Value]:
    async def evaluate(
        value: Value,
    ) -> EvaluatorResult:
        # Placeholder for the highest result
        highest: EvaluatorResult = EvaluatorResult(
            evaluator="highest",
            score=EvaluationScore(value=0),
            threshold=1,
            meta=None,
        )

        for result in await gather(
            evaluators(value),
            *[evaluator(value) for evaluator in _evaluators],
        ):
            if result.score >= highest.score:
                highest = result

        return highest

    return evaluate



from collections.abc import Iterable
from typing import Any

from haiway import ArgumentsTrace, ResultTrace, ctx

from draive.instructions import Instruction
from draive.lmm import (
    LMMCompletion,
    LMMContext,
    LMMContextElement,
    LMMInput,
    LMMInvocation,
    LMMOutput,
    LMMOutputSelection,
    LMMToolRequests,
    LMMToolResponses,
    LMMToolSelection,
    LMMToolSpecification,
)
from draive.metrics.tokens import TokenUsage
from draive.ollama.client import OllamaClient
from draive.ollama.config import OllamaChatConfig
from draive.ollama.models import ChatCompletionResponse, ChatMessage

__all__ = [
    "ollama_lmm",
]


def ollama_lmm(
    client: OllamaClient | None = None,
    /,
) -> LMMInvocation:
    client = client or OllamaClient.shared()

    async def lmm_invocation(
        *,
        instruction: Instruction | str | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        **extra: Any,
    ) -> LMMOutput:
        with ctx.scope("ollama_lmm_invocation"):
            ctx.record(
                ArgumentsTrace.of(
                    instruction=instruction,
                    context=context,
                    tool_selection=tool_selection,
                    tools=tools,
                    output=output,
                    **extra,
                )
            )

            config: OllamaChatConfig = ctx.state(OllamaChatConfig).updated(**extra)
            ctx.record(config)

            if tools:
                ctx.log_warning(
                    "Attempting to use Ollama with tools which is not supported."
                    " Ignoring provided tools..."
                )

            match output:
                case "auto" | "text":
                    config = config.updated(response_format="text")

                case "image":
                    raise NotImplementedError("image output is not supported by ollama")

                case "audio":
                    raise NotImplementedError("audio output is not supported by ollama")

                case "video":
                    raise NotImplementedError("video output is not supported by ollama")

                case _:
                    config = config.updated(response_format="json")

            messages: list[ChatMessage] = [
                _convert_context_element(element=element) for element in context
            ]

            if instruction:
                messages = [
                    ChatMessage(
                        role="system",
                        content=Instruction.of(instruction).format(),
                    ),
                    *messages,
                ]

            return await _chat_completion(
                client=client,
                config=config,
                messages=messages,
            )

    return LMMInvocation(invoke=lmm_invocation)


def _convert_context_element(
    element: LMMContextElement,
) -> ChatMessage:
    match element:
        case LMMInput() as input:
            return ChatMessage(
                role="user",
                content=input.content.as_string(),
            )

        case LMMCompletion() as completion:
            return ChatMessage(
                role="assistant",
                content=completion.content.as_string(),
            )

        case LMMToolRequests():
            raise NotImplementedError("Tools use is not supported by Ollama")

        case LMMToolResponses():
            raise NotImplementedError("Tools use is not supported by Ollama")


async def _chat_completion(
    *,
    client: OllamaClient,
    config: OllamaChatConfig,
    messages: list[ChatMessage],
) -> LMMOutput:
    completion: ChatCompletionResponse = await client.chat_completion(
        config=config,
        messages=messages,
    )

    ctx.record(
        TokenUsage.for_model(
            config.model,
            input_tokens=completion.prompt_eval_count,
            output_tokens=completion.eval_count,
        ),
    )

    ctx.record(ResultTrace.of(completion.message.content))
    return LMMCompletion.of(completion.message.content)



from draive.steps.call import steps_completion
from draive.steps.default import default_steps_completion
from draive.steps.state import Steps
from draive.steps.types import Step, StepsCompleting

__all__ = [
    "Step",
    "Steps",
    "StepsCompleting",
    "default_steps_completion",
    "steps_completion",
]



import json
from collections.abc import Mapping
from pathlib import Path
from typing import Any

from haiway import asynchronous

from draive.instructions.types import Instruction, InstructionFetching, MissingInstruction

__all__ = [
    "instructions_file",
]


def instructions_file(
    path: Path | str,
) -> InstructionFetching:
    storage: _InstructionsFileStorage = _InstructionsFileStorage(path=path)

    async def fetch(
        identifier: str,
        /,
        *,
        arguments: Mapping[str, str] | None = None,
        **extra: Any,
    ) -> Instruction:
        return await storage.instruction(
            identifier,
            arguments=arguments,
            **extra,
        )

    return fetch


class _InstructionsFileStorage:
    def __init__(
        self,
        path: Path | str,
    ) -> None:
        self._path: Path
        match path:
            case str() as path_str:
                self._path = Path(path_str)

            case path:
                self._path = path

        self._storage: Mapping[str, str] | None = None

    async def instruction(
        self,
        name: str,
        *,
        arguments: Mapping[str, str] | None = None,
        **extra: Any,
    ) -> Instruction:
        if self._storage is None:
            self._storage = await self._file_load()

        if instruction := self._storage.get(name):
            return Instruction.of(
                instruction,
                name=name,
                **(arguments if arguments is not None else {}),
            )

        else:
            raise MissingInstruction(
                "%s does not contain instruction for identifier %s",
                self._path,
                name,
            )

    @asynchronous
    def _file_load(
        self,
    ) -> Mapping[str, str]:
        if self._path.exists():
            with open(self._path, mode="rb") as file:
                match json.loads(file.read()):
                    case {**elements}:
                        return {
                            key: value for key, value in elements.items() if isinstance(value, str)
                        }

                    case _:
                        return {}

        else:
            return {}



from draive.mistral.client import MistralClient
from draive.mistral.config import MistralChatConfig, MistralEmbeddingConfig
from draive.mistral.embedding import mistral_text_embedding
from draive.mistral.lmm import mistral_lmm
from draive.mistral.tokenization import mistral_text_tokenizer
from draive.mistral.types import MistralException

__all__ = [
    "MistralChatConfig",
    "MistralClient",
    "MistralEmbeddingConfig",
    "MistralException",
    "mistral_lmm",
    "mistral_text_embedding",
    "mistral_text_tokenizer",
]



from collections.abc import Sequence

from haiway import State

from draive.conversation.default import default_conversation_completion
from draive.conversation.types import ConversationCompletion, ConversationMessage
from draive.utils import Memory

__all__ = [
    "Conversation",
]


class Conversation(State):
    completion: ConversationCompletion = default_conversation_completion
    memory: Memory[Sequence[ConversationMessage], ConversationMessage] | None = None



from collections.abc import AsyncIterator, Iterable
from typing import Any

from haiway import ctx

from draive.instructions import Instruction
from draive.lmm.state import LMMInvocation, LMMStream
from draive.lmm.types import (
    LMMContext,
    LMMOutput,
    LMMOutputSelection,
    LMMStreamInput,
    LMMStreamOutput,
    LMMStreamProperties,
    LMMToolSelection,
    LMMToolSpecification,
)
from draive.utils import ConstantStream

__all__ = [
    "lmm_invoke",
    "lmm_stream",
]


async def lmm_invoke(
    *,
    instruction: Instruction | str | None = None,
    context: LMMContext,
    tool_selection: LMMToolSelection = "auto",
    tools: Iterable[LMMToolSpecification] | None = None,
    output: LMMOutputSelection = "auto",
    **extra: Any,
) -> LMMOutput:
    return await ctx.state(LMMInvocation).invoke(
        instruction=instruction,
        context=context,
        tool_selection=tool_selection,
        tools=tools,
        output=output,
        **extra,
    )


async def lmm_stream(
    *,
    properties: AsyncIterator[LMMStreamProperties] | LMMStreamProperties,
    input: AsyncIterator[LMMStreamInput],  # noqa: A002
    context: LMMContext | None = None,
    **extra: Any,
) -> AsyncIterator[LMMStreamOutput]:
    properties_stream: AsyncIterator[LMMStreamProperties]
    match properties:
        case LMMStreamProperties() as constant:
            properties_stream = ConstantStream(constant)

        case variable:
            properties_stream = variable

    return await ctx.state(LMMStream).prepare(
        properties=properties_stream,
        input=input,
        context=context,
        **extra,
    )



from draive.generation.image.call import generate_image
from draive.generation.image.state import ImageGeneration
from draive.generation.image.typing import ImageGenerator

__all__ = [
    "ImageGeneration",
    "ImageGenerator",
    "generate_image",
]



from collections.abc import AsyncIterator, Iterable, Sequence
from typing import Any, Literal, overload

from haiway import ctx

from draive.conversation.state import Conversation
from draive.conversation.types import ConversationElement, ConversationMemory, ConversationMessage
from draive.instructions import Instruction
from draive.lmm import LMMStreamChunk
from draive.multimodal import Multimodal
from draive.prompts import Prompt
from draive.tools import AnyTool, Toolbox
from draive.utils import Memory

__all__ = [
    "conversation_completion",
]


@overload
async def conversation_completion(
    *,
    instruction: Instruction | str | None = None,
    input: ConversationMessage | Prompt | Multimodal,
    memory: ConversationMemory | Iterable[ConversationElement] | None = None,
    tools: Toolbox | Iterable[AnyTool] | None = None,
    stream: Literal[True],
    **extra: Any,
) -> AsyncIterator[LMMStreamChunk]: ...


@overload
async def conversation_completion(
    *,
    instruction: Instruction | str | None = None,
    input: ConversationMessage | Prompt | Multimodal,
    memory: ConversationMemory | Iterable[ConversationElement] | None = None,
    tools: Toolbox | Iterable[AnyTool] | None = None,
    stream: Literal[False] = False,
    **extra: Any,
) -> ConversationMessage: ...


async def conversation_completion(
    *,
    instruction: Instruction | str | None = None,
    input: ConversationMessage | Prompt | Multimodal,  # noqa: A002
    memory: ConversationMemory | Iterable[ConversationElement] | None = None,
    tools: Toolbox | Iterable[AnyTool] | None = None,
    stream: bool = False,
    **extra: Any,
) -> AsyncIterator[LMMStreamChunk] | ConversationMessage:
    conversation: Conversation = ctx.state(Conversation)

    # prepare memory
    conversation_memory: Memory[Sequence[ConversationMessage], ConversationMessage]
    match memory or conversation.memory:
        case None:
            conversation_memory = Memory[
                Sequence[ConversationMessage],
                ConversationMessage,
            ].constant(recalled=())

        case Memory() as memory:
            conversation_memory = memory

        case memory_messages:
            conversation_memory = Memory[
                Sequence[ConversationMessage],
                ConversationMessage,
            ].constant(recalled=tuple(memory_messages))

    # request completion
    return await conversation.completion(
        instruction=instruction,
        input=input,
        memory=conversation_memory,
        toolbox=Toolbox.out_of(tools),
        stream=stream,
    )



from collections.abc import Mapping, Sequence
from typing import Any, Literal, overload

from haiway import ctx

from draive.lmm import LMMInput
from draive.multimodal import MultimodalContent
from draive.prompts.state import PromptRepository
from draive.prompts.types import MissingPrompt, Prompt, PromptDeclaration

__all__ = [
    "fetch_prompt",
    "fetch_prompt_list",
]


async def fetch_prompt_list(
    **extra: Any,
) -> Sequence[PromptDeclaration]:
    return await ctx.state(PromptRepository).list(**extra)


@overload
async def fetch_prompt(
    reference: PromptDeclaration | str,
    /,
    *,
    default: Prompt | str | None = None,
    arguments: Mapping[str, str] | None = None,
    **extra: Any,
) -> Prompt | None: ...


@overload
async def fetch_prompt(
    reference: PromptDeclaration | str,
    /,
    *,
    default: Prompt | str,
    arguments: Mapping[str, str] | None = None,
    **extra: Any,
) -> Prompt: ...


@overload
async def fetch_prompt(
    reference: PromptDeclaration | str,
    /,
    *,
    default: Prompt | str | None = None,
    arguments: Mapping[str, str] | None = None,
    required: Literal[True],
    **extra: Any,
) -> Prompt: ...


async def fetch_prompt(
    reference: PromptDeclaration | str,
    /,
    *,
    default: Prompt | str | None = None,
    arguments: Mapping[str, str] | None = None,
    required: bool = True,
    **extra: Any,
) -> Prompt | None:
    name: str = reference if isinstance(reference, str) else reference.name

    match await ctx.state(PromptRepository).fetch(
        name,
        arguments=arguments,
        **extra,
    ):
        case None:
            match default:
                case None:
                    if required:
                        raise MissingPrompt(f"Missing prompt: '{name}'")

                    else:
                        return None

                case Prompt() as prompt:
                    return prompt

                case str() as text:
                    return Prompt.of(
                        LMMInput(
                            content=MultimodalContent.of(
                                text.format_map(arguments) if arguments else text
                            )
                        ),
                        name=name,
                        description=reference.description
                        if isinstance(reference, Prompt)
                        else None,
                    )

        case prompt:
            return prompt



from collections.abc import Mapping, Sequence
from typing import Any, Protocol, Self, final, overload, runtime_checkable
from uuid import uuid4

from haiway import State

from draive.parameters import DataModel, Field, ParameterSpecification

__all__ = [
    "Instruction",
    "InstructionDeclaration",
    "InstructionDeclarationArgument",
    "InstructionFetching",
    "MissingInstruction",
]


class MissingInstruction(Exception):
    pass


class InstructionDeclarationArgument(DataModel):
    name: str
    specification: ParameterSpecification = Field(
        specification={
            "type": "object",
            "additionalProperties": True,
        }
    )
    required: bool = True


class InstructionDeclaration(DataModel):
    name: str
    description: str | None = None
    arguments: Sequence[InstructionDeclarationArgument]


@final
class Instruction(State):
    @overload
    @classmethod
    def formatted(
        cls,
        instruction: None,
        /,
    ) -> None: ...

    @overload
    @classmethod
    def formatted(
        cls,
        instruction: Self | str,
        /,
        **variables: str,
    ) -> str: ...

    @overload
    @classmethod
    def formatted(
        cls,
        instruction: Self | str | None,
        /,
        **variables: str,
    ) -> str | None: ...

    @classmethod
    def formatted(
        cls,
        instruction: Self | str | None,
        /,
        **variables: str,
    ) -> str | None:
        match instruction:
            case None:
                return None

            case str() as string:
                if variables:
                    return string.format_map(variables)

                else:
                    return string

            case instruction:
                return instruction.format(**variables)

    @classmethod
    def of(
        cls,
        instruction: Self | str,
        /,
        *,
        name: str | None = None,
        description: str | None = None,
        **variables: str,
    ) -> Self:
        match instruction:
            case str() as content:
                return cls(
                    name=name or uuid4().hex,
                    description=description,
                    content=content,
                    variables=variables,
                )

            case instruction:
                return instruction.updated(**variables)

    name: str
    description: str | None
    content: str
    variables: Mapping[str, str]

    def format(
        self,
        **variables: str,
    ) -> str:
        if variables:
            return self.content.format_map(
                {
                    **self.variables,
                    **variables,
                },
            )

        elif self.variables:
            return self.content.format_map(self.variables)

        else:
            return self.content

    def extended(
        self,
        instruction: str,
        /,
        description: str | None = None,
        joiner: str | None = None,
        **variables: str,
    ) -> Self:
        if variables:
            return self.__class__(
                name=self.name,
                description=description,
                content=(joiner if joiner is not None else "").join((self.content, instruction)),
                variables={
                    **self.variables,
                    **variables,
                },
            )

        else:
            return self.__class__(
                name=self.name,
                description=description,
                content=(joiner if joiner is not None else "").join((self.content, instruction)),
                variables=self.variables,
            )

    def updated(
        self,
        **variables: str,
    ) -> Self:
        if variables:
            return self.__class__(
                name=self.name,
                description=self.description,
                content=self.content,
                variables={
                    **self.variables,
                    **variables,
                },
            )

        else:
            return self

    def __str__(self) -> str:
        try:
            return self.format()

        except KeyError:
            return self.content


@runtime_checkable
class InstructionFetching(Protocol):
    async def __call__(
        self,
        name: str,
        /,
        *,
        arguments: Mapping[str, str] | None = None,
        **extra: Any,
    ) -> Instruction | None: ...



from draive.conversation.call import conversation_completion
from draive.conversation.default import default_conversation_completion
from draive.conversation.state import Conversation
from draive.conversation.types import ConversationElement, ConversationMemory, ConversationMessage

__all__ = [
    "Conversation",
    "ConversationElement",
    "ConversationMemory",
    "ConversationMessage",
    "conversation_completion",
    "default_conversation_completion",
]



from typing import Any

from haiway import ctx

from draive.instructions import Instruction
from draive.multimodal import Multimodal, MultimodalContent
from draive.prompts import Prompt
from draive.steps.state import Steps
from draive.steps.types import Step

__all__ = [
    "steps_completion",
]


async def steps_completion(
    *steps: Step | Prompt | Multimodal,
    instruction: Instruction | str | None = None,
    **extra: Any,
) -> MultimodalContent:
    return await ctx.state(Steps).completion(
        *steps,
        instruction=instruction,
        **extra,
    )



from haiway import State

from draive.resources.types import ResourceFetching, ResourceListing

__all__ = [
    "ResourceRepository",
]


class ResourceRepository(State):
    list: ResourceListing
    fetch: ResourceFetching



from collections.abc import Iterable
from typing import Any, Literal

from haiway import ctx

from draive.generation.model.state import ModelGeneration
from draive.generation.model.types import ModelGeneratorDecoder
from draive.instructions import Instruction
from draive.multimodal import Multimodal
from draive.parameters import DataModel
from draive.prompts import Prompt
from draive.tools import AnyTool, Toolbox

__all__ = [
    "generate_model",
]


async def generate_model[Generated: DataModel](  # noqa: PLR0913
    generated: type[Generated],
    /,
    *,
    instruction: Instruction | str,
    input: Prompt | Multimodal,  # noqa: A002
    schema_injection: Literal["auto", "full", "simplified", "skip"] = "auto",
    tools: Toolbox | Iterable[AnyTool] | None = None,
    examples: Iterable[tuple[Multimodal, Generated]] | None = None,
    decoder: ModelGeneratorDecoder | None = None,
    **extra: Any,
) -> Generated:
    return await ctx.state(ModelGeneration).generate(
        generated,
        instruction=instruction,
        input=input,
        schema_injection=schema_injection,
        tools=tools,
        examples=examples,
        decoder=decoder,
        **extra,
    )



from collections.abc import Iterable, Sequence
from typing import Any

from haiway import ctx

from draive.instructions import Instruction
from draive.lmm import (
    LMMCompletion,
    LMMContextElement,
    LMMInput,
    LMMToolRequests,
    LMMToolResponse,
    LMMToolResponses,
    lmm_invoke,
)
from draive.multimodal import Multimodal, MultimodalContent
from draive.prompts import Prompt
from draive.tools import AnyTool, Toolbox

__all__ = [
    "default_generate_text",
]


async def default_generate_text(
    *,
    instruction: Instruction | str | None,
    input: Prompt | Multimodal,  # noqa: A002
    tools: Toolbox | Iterable[AnyTool] | None,
    examples: Iterable[tuple[Multimodal, str]] | None,
    **extra: Any,
) -> str:
    with ctx.scope("generate_text"):
        toolbox: Toolbox = Toolbox.out_of(tools)

        context: list[LMMContextElement] = [
            *[
                message
                for example in examples or []
                for message in [
                    LMMInput.of(example[0]),
                    LMMCompletion.of(example[1]),
                ]
            ],
        ]

        match input:
            case Prompt() as prompt:
                context.extend(prompt.content)

            case value:
                context.append(LMMInput.of(value))

        recursion_level: int = 0
        while recursion_level <= toolbox.repeated_calls_limit:
            match await lmm_invoke(
                instruction=instruction,
                context=context,
                tools=toolbox.available_tools(),
                tool_selection=toolbox.tool_selection(repetition_level=recursion_level),
                output="text",
                **extra,
            ):
                case LMMCompletion() as completion:
                    ctx.log_debug("Received text generation result")
                    return completion.content.as_string()

                case LMMToolRequests() as tool_requests:
                    ctx.log_debug("Received text generation tool calls")
                    responses: Sequence[LMMToolResponse] = await toolbox.respond_all(tool_requests)

                    if direct_responses := [response for response in responses if response.direct]:
                        return MultimodalContent.of(
                            *[response.content for response in direct_responses]
                        ).as_string()

                    else:
                        context.extend(
                            [
                                tool_requests,
                                LMMToolResponses(responses=responses),
                            ]
                        )

            recursion_level += 1  # continue with next recursion level

        raise RuntimeError("LMM exceeded limit of recursive calls")



from collections.abc import Sequence
from typing import Any

from haiway import ctx

from draive.instructions import Instruction
from draive.lmm import (
    LMMCompletion,
    LMMContextElement,
    LMMToolRequests,
    LMMToolResponse,
    LMMToolResponses,
    lmm_invoke,
)
from draive.multimodal import Multimodal, MultimodalContent
from draive.prompts import Prompt
from draive.steps.types import Step
from draive.tools import Toolbox

__all__ = [
    "default_steps_completion",
]


async def default_steps_completion(
    *steps: Step | Prompt | Multimodal,
    instruction: Instruction | str | None = None,
    **extra: Any,
) -> MultimodalContent:
    with ctx.scope("steps_completion"):
        assert steps, "Steps cannot be empty"  # nosec: B101
        result_parts: list[MultimodalContent] = []
        last_step_result: MultimodalContent | None = None
        context: list[LMMContextElement] = []
        volatile_context_range: range | None = None
        for step in steps:
            current_step: Step
            match step:
                case Step() as step:
                    current_step = step

                case input_content:
                    current_step = Step.of(input_content)

            if current_step.condition is not None and not current_step.condition():
                continue  # skip unavailable steps

            context_end_index: int = len(context)
            step_result: MultimodalContent = await _process_step(
                current_step,
                instruction=instruction,
                context=context,
                **extra,
            )

            # remove volatile parts of context after using it
            if volatile_context_range is not None:
                del context[volatile_context_range.start : volatile_context_range.stop]
                # adjust end index if removed stuff
                context_end_index -= volatile_context_range.stop - volatile_context_range.start
                volatile_context_range = None

            # mark new volatile parts of context if any to be removed
            if current_step.volatile:
                volatile_context_range = range(
                    context_end_index,
                    len(context),
                )

            # include step result as part of the final result if needed
            if current_step.extends_result:
                result_parts.append(step_result)
                last_step_result = None  # make sure we won't make duplicates

            else:
                last_step_result = step_result

        if last_step_result is not None:
            result_parts.append(last_step_result)

        return MultimodalContent.of(*result_parts)


async def _process_step(
    step: Step,
    /,
    *,
    instruction: Instruction | str | None,
    context: list[LMMContextElement],
    **extra: Any,
) -> MultimodalContent:
    match step.input:
        case Prompt() as prompt:
            context.extend(prompt.content)

        case input:
            context.append(input)

    if step.completion is not None:
        context.append(step.completion)
        return step.completion.content

    toolbox: Toolbox = step.toolbox

    recursion_level: int = 0
    context_end_index: int = len(context)
    while recursion_level <= toolbox.repeated_calls_limit:
        match await lmm_invoke(
            instruction=step.instruction or instruction,
            context=context,
            output=step.output,
            tools=toolbox.available_tools(),
            tool_selection=toolbox.tool_selection(repetition_level=recursion_level),
            **step.extra,
            **extra,
        ):
            case LMMCompletion() as completion:
                del context[context_end_index:]  # remove tool calls from context
                if step.result_processing is None:
                    context.append(completion)
                    return completion.content

                else:
                    processed_content: MultimodalContent = await step.result_processing(
                        completion.content
                    )
                    context.append(LMMCompletion.of(processed_content))
                    return processed_content

            case LMMToolRequests() as tool_requests:
                responses: Sequence[LMMToolResponse] = await toolbox.respond_all(tool_requests)

                if direct_results := [
                    response.content for response in responses if response.direct
                ]:
                    del context[context_end_index:]  # remove tool calls from context
                    direct_content: MultimodalContent = MultimodalContent.of(*direct_results)
                    if step.result_processing is None:
                        context.append(LMMCompletion.of(direct_content))
                        return direct_content

                    else:
                        processed_content: MultimodalContent = await step.result_processing(
                            direct_content
                        )
                        context.append(LMMCompletion.of(processed_content))
                        return processed_content

                else:
                    context.extend(
                        [
                            tool_requests,
                            LMMToolResponses(responses=responses),
                        ]
                    )

        recursion_level += 1  # continue with next recursion level

    raise RuntimeError("LMM exceeded limit of recursive calls")



from asyncio import Lock, gather
from collections.abc import Callable, Iterable, Mapping, Sequence
from pathlib import Path
from typing import Protocol, Self, overload, runtime_checkable
from uuid import UUID, uuid4

from haiway import asynchronous, ctx

from draive.evaluation.evaluator import EvaluatorResult, PreparedEvaluator
from draive.evaluation.generator import generate_case_parameters
from draive.evaluation.scenario import PreparedScenarioEvaluator, ScenarioEvaluatorResult
from draive.parameters import DataModel, Field

__all__ = [
    "EvaluationCaseResult",
    "EvaluationSuite",
    "EvaluationSuiteDefinition",
    "EvaluationSuiteStorage",
    "SuiteEvaluatorCaseResult",
    "SuiteEvaluatorResult",
    "evaluation_suite",
]


class EvaluationSuiteCase[CaseParameters: DataModel](DataModel):
    identifier: UUID = Field(default_factory=uuid4)
    parameters: CaseParameters
    comment: str | None = None


class SuiteEvaluatorCaseResult[CaseParameters: DataModel, Value: DataModel | str](DataModel):
    case: EvaluationSuiteCase[CaseParameters] = Field(
        description="Evaluated case",
    )
    value: Value = Field(
        description="Evaluated value",
    )
    results: Sequence[ScenarioEvaluatorResult] = Field(
        description="Evaluation results",
    )
    meta: Mapping[str, str | float | int | bool | None] | None = Field(
        description="Additional evaluation metadata",
        default=None,
    )

    @property
    def passed(self) -> bool:
        # empty results is equivalent of failure
        return len(self.results) > 0 and all(result.passed for result in self.results)

    def report(self) -> str:
        report: str = "\n---\n".join(
            [result.report() for result in self.results if not result.passed]
        )

        if report:  # nonempty report contains failing reports
            return (
                f"Evaluation case {self.case.identifier}:"
                f"\nvalue: {self.value}"
                f"\n{self.case.parameters}\n---\n{report}"
            )

        elif not self.results:
            return f"Evaluation case {self.case.identifier} empty!"

        else:
            return f"Evaluation case {self.case.identifier} passed!"


class SuiteEvaluatorResult[CaseParameters: DataModel, Value: DataModel | str](DataModel):
    cases: Sequence[SuiteEvaluatorCaseResult[CaseParameters, Value]]

    @property
    def passed(self) -> bool:
        return all(case.passed for case in self.cases)

    def report(self) -> str:
        report: str = "\n---\n".join(
            [result.report() for result in self.cases if not result.passed]
        )

        if report:  # nonempty report contains failing reports
            return f"Evaluation suite failed:\n\n{report}"

        elif not self.cases:
            return "Evaluation suite empty!"

        else:
            return "Evaluation suite passed!"


class EvaluationCaseResult[Value: DataModel | str](DataModel):
    @classmethod
    def of(
        cls,
        results: ScenarioEvaluatorResult | EvaluatorResult,
        *_results: ScenarioEvaluatorResult | EvaluatorResult,
        value: Value,
        meta: Mapping[str, str | float | int | bool | None] | None = None,
    ) -> Self:
        free_results: list[EvaluatorResult] = []
        scenario_results: list[ScenarioEvaluatorResult] = []
        for result in (results, *_results):
            match result:
                case ScenarioEvaluatorResult() as scenario_result:
                    scenario_results.append(scenario_result)

                case EvaluatorResult() as evaluator_result:
                    free_results.append(evaluator_result)

        if free_results:
            scenario_results.append(
                ScenarioEvaluatorResult(
                    name="EvaluationSuite",
                    evaluations=tuple(free_results),
                )
            )

        return cls(
            value=value,
            results=tuple(scenario_results),
            meta=meta,
        )

    @classmethod
    async def evaluating(
        cls,
        value: Value,
        /,
        evaluators: PreparedScenarioEvaluator[Value] | PreparedEvaluator[Value],
        *_evaluators: PreparedScenarioEvaluator[Value] | PreparedEvaluator[Value],
        meta: Mapping[str, str | float | int | bool | None] | None = None,
    ) -> Self:
        return cls.of(
            *await gather(
                *[evaluator(value) for evaluator in [evaluators, *_evaluators]],
                return_exceptions=False,
            ),
            value=value,
            meta=meta,
        )

    value: Value = Field(
        description="Evaluated value",
    )
    results: Sequence[ScenarioEvaluatorResult] = Field(
        description="Evaluation results",
    )
    meta: Mapping[str, str | float | int | bool | None] | None = Field(
        description="Additional evaluation metadata",
        default=None,
    )


@runtime_checkable
class EvaluationSuiteDefinition[CaseParameters: DataModel, Value: DataModel | str](Protocol):
    async def __call__(
        self,
        parameters: CaseParameters,
    ) -> EvaluationCaseResult[Value]: ...


class EvaluationSuiteData[CaseParameters: DataModel](DataModel):
    cases: Sequence[EvaluationSuiteCase[CaseParameters]] = Field(default_factory=tuple)


@runtime_checkable
class EvaluationSuiteStorage[CaseParameters: DataModel](Protocol):
    async def load(
        self,
    ) -> EvaluationSuiteData[CaseParameters]: ...

    async def save(
        self,
        data: EvaluationSuiteData[CaseParameters],
    ) -> None: ...


class EvaluationSuite[CaseParameters: DataModel, Value: DataModel | str]:
    def __init__(
        self,
        definition: EvaluationSuiteDefinition[CaseParameters, Value],
        storage: EvaluationSuiteStorage[CaseParameters],
        parameters: type[CaseParameters],
    ) -> None:
        self._definition: EvaluationSuiteDefinition[CaseParameters, Value] = definition
        self._storage: EvaluationSuiteStorage[CaseParameters] = storage
        self._data_cache: EvaluationSuiteData[CaseParameters] | None = None
        self._parameters: type[CaseParameters] = parameters
        self._lock: Lock = Lock()

    @overload
    async def __call__(
        self,
        parameters: CaseParameters | UUID | None,
        /,
        *,
        reload: bool = False,
    ) -> SuiteEvaluatorCaseResult[CaseParameters, Value]: ...

    @overload
    async def __call__(
        self,
        /,
        *,
        reload: bool = False,
    ) -> SuiteEvaluatorResult[CaseParameters, Value]: ...

    async def __call__(
        self,
        parameters: CaseParameters | UUID | None = None,
        /,
        *,
        reload: bool = False,
    ) -> (
        SuiteEvaluatorResult[CaseParameters, Value]
        | SuiteEvaluatorCaseResult[CaseParameters, Value]
    ):
        async with self._lock:
            match parameters:
                case None:
                    return SuiteEvaluatorResult(
                        cases=await gather(
                            *[
                                self._evaluate(case=case)
                                for case in (await self._data(reload=reload)).cases
                            ],
                            return_exceptions=False,
                        )
                    )

                case UUID() as identifier:
                    available_cases: Sequence[EvaluationSuiteCase[CaseParameters]] = (
                        await self._data(reload=reload)
                    ).cases

                    if evaluation_case := next(
                        iter([case for case in available_cases if case.identifier == identifier]),
                        None,
                    ):
                        return await self._evaluate(case=evaluation_case)

                    else:
                        raise ValueError(f"Evaluation case with ID {identifier} does not exists.")

                case case_parameters:
                    return await self._evaluate(
                        case=EvaluationSuiteCase[CaseParameters](
                            parameters=case_parameters,
                        )
                    )

    async def _evaluate(
        self,
        *,
        case: EvaluationSuiteCase[CaseParameters],
    ) -> SuiteEvaluatorCaseResult[CaseParameters, Value]:
        result: EvaluationCaseResult[Value] = await self._definition(parameters=case.parameters)

        return SuiteEvaluatorCaseResult[CaseParameters, Value](
            case=case,
            value=result.value,
            results=result.results,
        )

    async def _data(
        self,
        reload: bool = False,
    ) -> EvaluationSuiteData[CaseParameters]:
        if reload or self._data_cache is None:
            self._data_cache = await self._storage.load()
            return self._data_cache

        else:
            return self._data_cache

    async def cases(
        self,
        reload: bool = False,
    ) -> Sequence[EvaluationSuiteCase[CaseParameters]]:
        async with self._lock:
            return (await self._data(reload=reload)).cases

    async def add_case(
        self,
        parameters: CaseParameters,
        /,
        comment: str | None = None,
    ) -> None:
        async with self._lock:
            data: EvaluationSuiteData[CaseParameters] = await self._data(reload=True)
            self._data_cache = data.updated(
                cases=(
                    *data.cases,
                    EvaluationSuiteCase[CaseParameters](
                        parameters=parameters,
                        comment=comment,
                    ),
                )
            )
            await self._storage.save(self._data_cache)

    async def generate_cases(
        self,
        *,
        persist: bool = False,
        count: int,
        examples: Iterable[CaseParameters] | None = None,
    ) -> None:
        async with self._lock:
            data: EvaluationSuiteData[CaseParameters] = await self._data(reload=True)
            self._data_cache = data.updated(
                cases=(
                    *data.cases,
                    *[
                        EvaluationSuiteCase(parameters=parameters)
                        for parameters in await generate_case_parameters(
                            self._parameters,
                            count=count,
                            examples=examples or [case.parameters for case in data.cases],
                        )
                    ],
                )
            )

            if persist:
                await self._storage.save(self._data_cache)

    async def remove_case(
        self,
        identifier: UUID,
        /,
    ) -> None:
        async with self._lock:
            data: EvaluationSuiteData[CaseParameters] = await self._data(reload=True)
            self._data_cache = data.updated(
                cases=tuple(case for case in data.cases if case.identifier != identifier)
            )
            await self._storage.save(self._data_cache)


def evaluation_suite[CaseParameters: DataModel, Value: DataModel | str](
    case: type[CaseParameters],
    /,
    storage: EvaluationSuiteStorage[CaseParameters] | Path | str | None = None,
) -> Callable[
    [EvaluationSuiteDefinition[CaseParameters, Value]],
    EvaluationSuite[CaseParameters, Value],
]:
    suite_storage: EvaluationSuiteStorage[CaseParameters]
    match storage:
        case None:
            suite_storage = _EvaluationSuiteMemoryStorage[CaseParameters](
                data_type=EvaluationSuiteData[case],
            )

        case str() as path_str:
            suite_storage = _EvaluationSuiteFileStorage[CaseParameters](
                path=path_str,
                data_type=EvaluationSuiteData[case],
            )

        case Path() as path:
            suite_storage = _EvaluationSuiteFileStorage[CaseParameters](
                path=path,
                data_type=EvaluationSuiteData[case],
            )

        case storage:
            suite_storage = storage

    def wrap(
        definition: EvaluationSuiteDefinition[CaseParameters, Value],
    ) -> EvaluationSuite[CaseParameters, Value]:
        return EvaluationSuite[CaseParameters, Value](
            definition=definition,
            storage=suite_storage,
            parameters=case,
        )

    return wrap


class _EvaluationSuiteMemoryStorage[CaseParameters: DataModel]:
    def __init__(
        self,
        data_type: type[EvaluationSuiteData[CaseParameters]],
    ) -> None:
        self.store: Sequence[EvaluationSuiteCase[CaseParameters]] = ()
        self._data_type: type[EvaluationSuiteData[CaseParameters]] = data_type

    async def load(
        self,
    ) -> EvaluationSuiteData[CaseParameters]:
        return self._data_type(cases=self.store)

    async def save(
        self,
        data: EvaluationSuiteData[CaseParameters],
    ) -> None:
        self.store = data.cases


class _EvaluationSuiteFileStorage[CaseParameters: DataModel]:
    def __init__(
        self,
        path: Path | str,
        data_type: type[EvaluationSuiteData[CaseParameters]],
    ) -> None:
        self._path: Path
        match path:
            case str() as path_str:
                self._path = Path(path_str)

            case path:
                self._path = path

        self._data_type: type[EvaluationSuiteData[CaseParameters]] = data_type

    async def load(
        self,
    ) -> EvaluationSuiteData[CaseParameters]:
        try:
            return await self._file_load(data_type=self._data_type)

        except ValueError as exc:
            ctx.log_error(
                f"Invalid EvaluationSuite at {self._path}",
                exception=exc,
            )
            return self._data_type()

    @asynchronous
    def _file_load(
        self,
        data_type: type[EvaluationSuiteData[CaseParameters]],
    ) -> EvaluationSuiteData[CaseParameters]:
        if self._path.exists():
            with open(self._path, mode="rb") as file:
                return data_type.from_json(file.read())

        else:
            return data_type()

    async def save(
        self,
        data: EvaluationSuiteData[CaseParameters],
    ) -> None:
        await self._file_save(data=data)

    @asynchronous
    def _file_save(
        self,
        data: EvaluationSuiteData[CaseParameters],
    ) -> None:
        with open(self._path, mode="wb+") as file:
            file.write(data.as_json(indent=2).encode("utf-8"))



from asyncio import gather
from typing import Any

from haiway import ctx, not_missing
from openai.types.moderation import Moderation

from draive.multimodal import MultimodalContent
from draive.openai.client import OpenAIClient
from draive.openai.config import OpenAIModerationConfig
from draive.safeguards import ContentGuardrails, GuardrailsException
from draive.splitters import split_text

__all__ = [
    "openai_content_guardrails",
]


async def openai_content_guardrails(  # noqa: C901, PLR0915
    client: OpenAIClient | None = None,
    /,
) -> ContentGuardrails:
    client = client or OpenAIClient.shared()

    async def openai_content_guardrails(  # noqa: C901, PLR0912
        content: MultimodalContent,
        /,
        **extra: Any,
    ) -> MultimodalContent:
        config: OpenAIModerationConfig = ctx.state(OpenAIModerationConfig).updated(**extra)

        content_parts: list[str] = split_text(
            content.as_string(),
            part_size=2000,
            part_overlap_size=200,
            count_size=len,
        )
        results: list[Moderation] = await gather(
            *[client.moderation_check(text=part) for part in content_parts]
        )

        violations: set[str] = set()
        for result in results:
            if (
                not_missing(config.harassment_threshold)
                and result.category_scores.harassment >= config.harassment_threshold
            ):
                violations.add("harassment")

            elif result.categories.harassment:
                violations.add("harassment")

            if (
                not_missing(config.harassment_threatening_threshold)
                and result.category_scores.harassment_threatening
                >= config.harassment_threatening_threshold
            ):
                violations.add("harassment_threatening")

            elif result.categories.harassment_threatening:
                violations.add("harassment_threatening")

            if (
                not_missing(config.hate_threshold)
                and result.category_scores.hate >= config.hate_threshold
            ):
                violations.add("hate")

            elif result.categories.hate:
                violations.add("hate")

            if (
                not_missing(config.hate_threatening_threshold)
                and result.category_scores.hate_threatening >= config.hate_threatening_threshold
            ):
                violations.add("hate_threatening")

            elif result.categories.hate_threatening:
                violations.add("hate_threatening")

            if (
                not_missing(config.self_harm_threshold)
                and result.category_scores.self_harm >= config.self_harm_threshold
            ):
                violations.add("self_harm")

            elif result.categories.self_harm:
                violations.add("self_harm")

            if (
                not_missing(config.self_harm_instructions_threshold)
                and result.category_scores.self_harm_instructions
                >= config.self_harm_instructions_threshold
            ):
                violations.add("self_harm_instructions")

            elif result.categories.self_harm_instructions:
                violations.add("self_harm_instructions")

            if (
                not_missing(config.self_harm_intent_threshold)
                and result.category_scores.self_harm_intent >= config.self_harm_intent_threshold
            ):
                violations.add("self_harm_intent")

            elif result.categories.self_harm_intent:
                violations.add("self_harm_intent")

            if (
                not_missing(config.sexual_threshold)
                and result.category_scores.sexual >= config.sexual_threshold
            ):
                violations.add("sexual")

            elif result.categories.sexual:
                violations.add("sexual")

            if (
                not_missing(config.sexual_minors_threshold)
                and result.category_scores.sexual_minors >= config.sexual_minors_threshold
            ):
                violations.add("sexual_minors")

            elif result.categories.sexual_minors:
                violations.add("sexual_minors")

            if (
                not_missing(config.violence_threshold)
                and result.category_scores.violence >= config.violence_threshold
            ):
                violations.add("violence")

            elif result.categories.violence:
                violations.add("violence")

            if (
                not_missing(config.violence_graphic_threshold)
                and result.category_scores.violence_graphic >= config.violence_graphic_threshold
            ):
                violations.add("violence_graphic")

            elif result.categories.violence_graphic:
                violations.add("violence_graphic")

        if violations:
            raise GuardrailsException(f"Content violated rule(s): {violations}")

        return content  # no violation - use content

    return openai_content_guardrails



from pathlib import Path

from draive.sentencepiece import sentencepiece_processor, sentencepiece_tokenizer
from draive.tokenization import Tokenization

__all__ = [
    "mistral_text_tokenizer",
]


async def mistral_text_tokenizer(
    model_name: str,
    /,
) -> Tokenization:
    return Tokenization(
        tokenize_text=sentencepiece_tokenizer(
            await sentencepiece_processor(model_path=_model_path(model_name))
        )
    )


def _model_path(
    model_name: str,
    /,
) -> str:
    model_file: str = _mapping.get(model_name, "mistral_v3_tokenizer.model")
    return str(Path(__file__).parent / "tokens" / model_file)


_mapping: dict[str, str] = {
    "open-mistral-7b": "mistral_v1_tokenizer.model",
    "open-mixtral-8x7b": "mistral_v1_tokenizer.model",
    "mistral-embed": "mistral_v1_tokenizer.model",
    "mistral-small": "mistral_v2_tokenizer.model",
    "mistral-large": "mistral_v2_tokenizer.model",
    "open-mixtral-8x22b": "mistral_v3_tokenizer.model",
    "codestral-22b": "mistral_v3_tokenizer.model",
}



from draive.mcp.client import MCPClient
from draive.mcp.server import expose_prompts, expose_resources, expose_tools

__all__ = [
    "MCPClient",
    "expose_prompts",
    "expose_resources",
    "expose_tools",
]



from asyncio import gather
from collections.abc import Iterable, Mapping, Sequence
from typing import Any, Literal, Protocol, Self, cast, final, runtime_checkable

from haiway import State, ctx

from draive.lmm.types import (
    LMMToolError,
    LMMToolException,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponse,
    LMMToolSpecification,
)
from draive.multimodal import MultimodalContent
from draive.tools.tool import AnyTool

__all__ = [
    "ExternalToolbox",
    "Toolbox",
    "ToolboxFetching",
]


@final
class Toolbox(State):
    @classmethod
    def of(
        cls,
        *tools: AnyTool,
        suggest: AnyTool | bool | None = None,
        repeated_calls_limit: int | None = None,
    ) -> Self:
        match suggest:
            case None | False:
                return cls(
                    tools={tool.name: tool for tool in tools},
                    suggest_tool=False,
                    repeated_calls_limit=repeated_calls_limit or 1,
                )

            case True:
                return cls(
                    tools={tool.name: tool for tool in tools},
                    suggest_tool=True,
                    repeated_calls_limit=repeated_calls_limit or 1,
                )

            case tool:
                return cls(
                    tools={**{tool.name: tool for tool in tools}, tool.name: tool},
                    suggest_tool=tool,
                    repeated_calls_limit=repeated_calls_limit or 1,
                )

    @classmethod
    def out_of(
        cls,
        tools: Self | Iterable[AnyTool] | None,
        /,
    ) -> Self:
        match tools:
            case None:
                return cls(
                    tools={},
                    suggest_tool=False,
                    repeated_calls_limit=1,
                )

            case Toolbox() as toolbox:
                return toolbox

            case tools:
                return cls(
                    tools={tool.name: tool for tool in tools},
                    suggest_tool=False,
                    repeated_calls_limit=1,
                )

    @classmethod
    async def external(
        cls,
        suggest: bool | None = None,
        repeated_calls_limit: int | None = None,
        other_tools: Iterable[AnyTool] | None = None,
        **extra: Any,
    ) -> Self:
        external_toolbox: Self = cast(
            Self,
            await ctx.state(ExternalToolbox).fetch(
                suggest_tool=suggest,
                repeated_calls_limit=repeated_calls_limit,
                **extra,
            ),
        )

        if other_tools:
            return external_toolbox.updated(
                tools={**external_toolbox.tools, **{tool.name: tool for tool in other_tools}}
            )

        else:
            return external_toolbox

    tools: Mapping[str, AnyTool]
    suggest_tool: AnyTool | bool
    repeated_calls_limit: int

    def tool_selection(
        self,
        *,
        repetition_level: int = 0,
    ) -> LMMToolSpecification | Literal["auto", "required", "none"]:
        if repetition_level >= self.repeated_calls_limit:
            return "none"  # require no tools if reached the limit

        elif repetition_level != 0:
            return "auto"  # require tools only for the first call, use auto otherwise

        elif self.suggest_tool is False:
            return "auto"

        elif self.suggest_tool is True:
            return "required"

        elif self.suggest_tool.available:
            return self.suggest_tool.specification  # use suggested tool if able

        else:
            return "auto"

    def available_tools(self) -> Sequence[LMMToolSpecification]:
        return [tool.specification for tool in self.tools.values() if tool.available]

    async def call_tool(
        self,
        name: str,
        /,
        call_id: str,
        arguments: Mapping[str, Any],
    ) -> MultimodalContent:
        if tool := self.tools.get(name):
            return await tool._toolbox_call(  # pyright: ignore[reportPrivateUsage]
                call_id,
                arguments=arguments,
            )

        else:
            raise LMMToolException("Requested tool (%s) is not defined", name)

    async def respond_all(
        self,
        requests: LMMToolRequests,
        /,
    ) -> Sequence[LMMToolResponse]:
        return await gather(
            *[self.respond(request) for request in requests.requests],
            return_exceptions=False,  # toolbox calls handle errors if able
        )

    async def respond(
        self,
        request: LMMToolRequest,
        /,
    ) -> LMMToolResponse:
        if tool := self.tools.get(request.tool):
            try:
                return LMMToolResponse(
                    identifier=request.identifier,
                    tool=request.tool,
                    content=await tool._toolbox_call(  # pyright: ignore[reportPrivateUsage]
                        request.identifier,
                        arguments=request.arguments or {},
                    ),
                    direct=tool.requires_direct_result,
                    error=False,
                )

            except LMMToolError as error:  # use formatted error, blow up on other exception
                ctx.log_error(
                    "Tool (%s) returned an error",
                    request.tool,
                    exception=error,
                )
                return LMMToolResponse(
                    identifier=request.identifier,
                    tool=request.tool,
                    content=error.content,
                    direct=False,
                    error=True,
                )

        else:
            # log error and provide fallback result to avoid blowing out the execution
            ctx.log_error("Requested tool (%s) is not defined", request.tool)
            return LMMToolResponse(
                identifier=request.identifier,
                tool=request.tool,
                content=MultimodalContent.of("ERROR"),
                direct=False,
                error=True,
            )


@runtime_checkable
class ToolboxFetching(Protocol):
    async def __call__(
        self,
        *,
        suggest: bool | None = None,
        repeated_calls_limit: int | None = None,
        **extra: Any,
    ) -> Toolbox: ...


class ExternalToolbox(State):
    fetch: ToolboxFetching



from base64 import b64decode, b64encode
from collections.abc import Mapping, Sequence
from typing import Final, Literal, Self, cast, get_args

from draive.multimodal.data_field import b64_or_url_field
from draive.parameters import DataModel

__all__ = [
    "MEDIA_KINDS",
    "MEDIA_TYPES",
    "MediaContent",
    "MediaKind",
    "MediaType",
    "validated_media_kind",
    "validated_media_type",
]

MediaType = Literal[
    "image/jpeg",
    "image/png",
    "image/bmp",
    "image/gif",
    "audio/aac",
    "audio/mpeg",
    "audio/ogg",
    "audio/wav",
    "video/mp4",
    "video/mpeg",
    "video/ogg",
]
MEDIA_TYPES: Final[Sequence[str]] = get_args(MediaType)


def validated_media_type(
    mime: str,
    /,
) -> MediaType:
    if mime in MEDIA_TYPES:
        return cast(MediaType, mime)

    else:
        raise ValueError(f"Unsupported media type: {mime}")


MediaKind = Literal[
    "image",
    "audio",
    "video",
]
MEDIA_KINDS: Final[Sequence[str]] = get_args(MediaKind)


def validated_media_kind(
    kind: str,
    /,
) -> MediaKind:
    if kind in MEDIA_KINDS:
        return cast(MediaKind, kind)

    else:
        raise ValueError(f"Unsupported media kind: {kind}")


class MediaContent(DataModel):
    @classmethod
    def url(
        cls,
        url: str,
        /,
        media: MediaType | MediaKind,
        meta: Mapping[str, str | float | int | bool | None] | None = None,
    ) -> Self:
        return cls(
            media=media,
            source=url,
            meta=meta,
        )

    @classmethod
    def base64(
        cls,
        data: str,
        /,
        media: MediaType,
        meta: Mapping[str, str | float | int | bool | None] | None = None,
    ) -> Self:
        return cls(
            media=media,
            source=b64decode(data),
            meta=meta,
        )

    @classmethod
    def data(
        cls,
        data: bytes,
        /,
        media: MediaType,
        meta: Mapping[str, str | float | int | bool | None] | None = None,
    ) -> Self:
        return cls(
            media=media,
            source=data,
            meta=meta,
        )

    media: MediaType | MediaKind
    # special field - url string or base64 content auto converted to bytes
    source: str | bytes = b64_or_url_field()
    meta: Mapping[str, str | float | int | bool | None] | None = None

    @property
    def kind(self) -> MediaKind:
        match self.media:
            case "image" | "image/jpeg" | "image/png" | "image/bmp" | "image/gif":
                return "image"

            case "audio" | "audio/aac" | "audio/mpeg" | "audio/ogg" | "audio/wav":
                return "audio"

            case "video" | "video/mp4" | "video/mpeg" | "video/ogg":
                return "video"

    def as_string(
        self,
        *,
        include_data: bool,
    ) -> str:
        match self.source:
            case str() as string:
                return f"![{self.kind}]({string})"

            case bytes() as data:
                if include_data:
                    return f"![{self.kind}](data:{self.media};base64,{b64encode(data).decode()})"

                else:
                    return f"![{self.kind}]()"

    def __bool__(self) -> bool:
        return len(self.source) > 0



from draive.gemini.client import GeminiClient
from draive.gemini.config import GeminiConfig, GeminiEmbeddingConfig
from draive.gemini.embedding import gemini_text_embedding
from draive.gemini.lmm import gemini_lmm
from draive.gemini.tokenization import gemini_tokenizer
from draive.gemini.types import GeminiException

__all__ = [
    "GeminiClient",
    "GeminiConfig",
    "GeminiEmbeddingConfig",
    "GeminiException",
    "gemini_lmm",
    "gemini_text_embedding",
    "gemini_tokenizer",
]



from base64 import b64encode
from collections.abc import Iterable, Mapping, Sequence
from typing import Any
from uuid import uuid4

from haiway import as_dict
from mcp.server import Server
from mcp.types import EmbeddedResource, GetPromptResult
from mcp.types import ImageContent as MCPImageContent
from mcp.types import Prompt as MCPPrompt
from mcp.types import PromptArgument as MCPPromptArgument
from mcp.types import PromptMessage as MCPPromptMessage
from mcp.types import Resource as MCPResource
from mcp.types import TextContent as MCPTextContent
from mcp.types import Tool as MCPTool
from pydantic import AnyUrl

from draive.lmm import LMMCompletion, LMMContextElement, LMMInput
from draive.multimodal import MediaContent, MultimodalContent, TextContent
from draive.prompts import Prompt, PromptTemplate
from draive.resources import Resource, ResourceContent, ResourceTemplate
from draive.tools import AnyTool, Toolbox

__all__ = [
    "expose_prompts",
    "expose_resources",
    "expose_tools",
]


def expose_resources(  # noqa: C901
    resources: Iterable[ResourceTemplate[Any, Any] | Resource],
    /,
    server: Server,
) -> None:
    resource_declarations: list[MCPResource] = []
    available_resources: dict[str, ResourceTemplate[Any, Any] | Resource] = {}
    for resource in resources:
        match resource:
            case Resource():
                available_resources[resource.uri] = resource
                match resource.content:
                    case ResourceContent() as content:
                        resource_declarations.append(
                            MCPResource(
                                uri=AnyUrl(resource.uri),
                                mimeType=content.mime_type,
                                name=resource.name,
                                description=resource.description,
                            )
                        )

                    case _:
                        raise NotImplementedError("Multi-content resources are not supported yet")

            case ResourceTemplate():
                resource_declarations.append(
                    MCPResource(
                        uri=AnyUrl(resource.uri),
                        mimeType=resource.declaration.mime_type,
                        name=resource.declaration.name,
                        description=resource.declaration.description,
                    )
                )
                available_resources[resource.uri] = resource

    @server.list_resources()
    async def list_resources() -> list[MCPResource]:  # pyright: ignore[reportUnusedFunction]
        return resource_declarations

    @server.read_resource()
    async def read_resource(uri: AnyUrl) -> str | bytes:  # pyright: ignore[reportUnusedFunction]
        resource: Resource
        match available_resources.get(uri.unicode_string()):
            case None:
                raise ValueError(f"Resource '{uri}' is not defined")

            case Resource() as available_resource:
                resource = available_resource

            case ResourceTemplate() as resource_template:
                # TODO: we can support only resource templates without actual arguments?
                # perhaps we could extract uri arguments
                resource = await resource_template.resolve()

        match resource.content:
            case ResourceContent() as content:
                return content.blob

            case _:
                raise NotImplementedError("Multi-content resources are not supported yet")


def expose_prompts(
    prompts: Iterable[PromptTemplate[Any] | Prompt],
    /,
    server: Server,
) -> None:
    prompt_declarations: list[MCPPrompt] = []
    available_prompts: dict[str, PromptTemplate[Any] | Prompt] = {}
    for prompt in prompts:
        match prompt:
            case Prompt():
                prompt_declarations.append(
                    MCPPrompt(
                        name=prompt.name,
                        arguments=None,
                        description=prompt.description,
                    )
                )
                available_prompts[prompt.name] = prompt

            case PromptTemplate():
                prompt_declarations.append(
                    MCPPrompt(
                        name=prompt.declaration.name,
                        arguments=[
                            MCPPromptArgument(
                                name=argument.name,
                                description=argument.specification.get("description"),
                                required=argument.required,
                            )
                            for argument in prompt.declaration.arguments
                        ],
                        description=prompt.declaration.description,
                    )
                )
                available_prompts[prompt.declaration.name] = prompt

    @server.list_prompts()
    async def list_prompts() -> list[MCPPrompt]:  # pyright: ignore[reportUnusedFunction]
        return prompt_declarations

    @server.get_prompt()
    async def get_prompt(  # pyright: ignore[reportUnusedFunction]
        name: str,
        arguments: Mapping[str, Any] | None,
    ) -> GetPromptResult:
        match available_prompts.get(name):
            case None:
                raise ValueError(f"Prompt '{name}' is not defined")

            case Prompt() as direct_prompt:
                return GetPromptResult(
                    description=direct_prompt.description,
                    messages=[_convert_message(element) for element in direct_prompt.content],
                )

            case PromptTemplate() as dynamic_prompt:
                resolved_prompt: Prompt = await dynamic_prompt.resolve(arguments or {})
                return GetPromptResult(
                    description=resolved_prompt.description,
                    messages=[_convert_message(element) for element in resolved_prompt.content],
                )


def _convert_message(
    element: LMMContextElement,
    /,
) -> MCPPromptMessage:
    match element:
        case LMMInput():
            match _convert_multimodal_content(element.content):
                case [content]:
                    return MCPPromptMessage(
                        role="user",
                        content=content,
                    )
                case []:
                    return MCPPromptMessage(
                        role="user",
                        # convert empty content to empty text to prevent errors
                        content=MCPTextContent(
                            type="text",
                            text="",
                        ),
                    )

                case [*_]:
                    raise NotImplementedError(
                        f"Unsuppoprted MCP prompt element content: {element.content}!"
                    )

        case LMMCompletion():
            match _convert_multimodal_content(element.content):
                case [content]:
                    return MCPPromptMessage(
                        role="assistant",
                        content=content,
                    )
                case []:
                    return MCPPromptMessage(
                        role="assistant",
                        # convert empty content to empty text to prevent errors
                        content=MCPTextContent(
                            type="text",
                            text="",
                        ),
                    )

                case [*_]:
                    raise NotImplementedError(
                        f"Unsuppoprted MCP prompt element content: {element.content}!"
                    )

        case _:
            raise NotImplementedError(f"Unsuppoprted MCP prompt context element: {type(element)}!")


def expose_tools(
    tools: Toolbox | Iterable[AnyTool],
    /,
    server: Server,
) -> None:
    toolbox: Toolbox = Toolbox.out_of(tools)

    @server.list_tools()
    async def list_tools() -> list[MCPTool]:  # pyright: ignore[reportUnusedFunction]
        return [
            MCPTool(
                name=tool["name"],
                description=tool["description"],
                inputSchema=as_dict(tool["parameters"]),
            )
            for tool in toolbox.available_tools()
        ]

    @server.call_tool()
    async def call_tool(  # pyright: ignore[reportUnusedFunction]
        name: str,
        arguments: Mapping[str, Any],
    ) -> Sequence[MCPTextContent | MCPImageContent | EmbeddedResource]:
        return _convert_multimodal_content(
            await toolbox.call_tool(
                name,
                call_id=uuid4().hex,
                arguments=arguments,
            )
        )


def _convert_multimodal_content(
    content: MultimodalContent,
) -> Sequence[MCPTextContent | MCPImageContent | EmbeddedResource]:
    converted: list[MCPTextContent | MCPImageContent | EmbeddedResource] = []
    for part in content.parts:
        match part:
            case TextContent() as text:
                converted.append(
                    MCPTextContent(
                        type="text",
                        text=text.text,
                    )
                )

            case MediaContent() as media:
                if media.kind != "image":
                    raise NotImplementedError(f"Media support for {media.media} is not implemented")

                if not isinstance(media.source, bytes):
                    # TODO: download images on the fly?
                    raise NotImplementedError(
                        "Media support for urls is not implemented, please use data blobs instead"
                    )

                converted.append(
                    MCPImageContent(
                        type="image",
                        data=b64encode(media.source).decode(),
                        mimeType=media.media,
                    )
                )

            case other:
                converted.append(
                    MCPTextContent(
                        type="text",
                        text=other.as_json(),
                    )
                )

    return converted



from haiway import State

from draive.prompts.types import PromptFetching, PromptListing

__all__ = [
    "PromptRepository",
]


class PromptRepository(State):
    list: PromptListing
    fetch: PromptFetching



from collections.abc import Iterable
from typing import Any

from haiway import ctx

from draive.generation.text.state import TextGeneration
from draive.instructions import Instruction
from draive.multimodal import Multimodal
from draive.prompts import Prompt
from draive.tools import AnyTool, Toolbox

__all__ = [
    "generate_text",
]


async def generate_text(
    *,
    instruction: Instruction | str | None = None,
    input: Prompt | Multimodal,  # noqa: A002
    tools: Toolbox | Iterable[AnyTool] | None = None,
    examples: Iterable[tuple[Multimodal, str]] | None = None,
    **extra: Any,
) -> str:
    return await ctx.state(TextGeneration).generate(
        instruction=instruction,
        input=input,
        tools=tools,
        examples=examples,
        **extra,
    )



from collections.abc import Callable, Coroutine, Sequence
from typing import Protocol, final

from haiway import ArgumentsTrace, ResultTrace, ctx, freeze

from draive.parameters import ParametrizedFunction
from draive.resources.types import Resource, ResourceContent, ResourceDeclaration

__all__ = [
    "ResourceAvailabilityCheck",
    "ResourceTemplate",
    "resource",
]


class ResourceAvailabilityCheck(Protocol):
    def __call__(self) -> bool: ...


# TODO: add https://modelcontextprotocol.io/docs/concepts/resources#resource-templates
# support and uri template resolving
@final
class ResourceTemplate[**Args, Result: Sequence[Resource] | ResourceContent](
    ParametrizedFunction[Args, Coroutine[None, None, Result]]
):
    def __init__(  # noqa: PLR0913
        self,
        /,
        uri: str,
        *,
        mime_type: str | None,
        name: str,
        description: str | None,
        availability_check: ResourceAvailabilityCheck | None = None,
        function: Callable[Args, Coroutine[None, None, Result]],
    ) -> None:
        super().__init__(function)

        self.uri: str = uri
        self.declaration: ResourceDeclaration = ResourceDeclaration(
            uri=uri,
            mime_type=mime_type,
            name=name,
            description=description,
        )
        self._check_availability: ResourceAvailabilityCheck = availability_check or (
            lambda: True  # available by default
        )

        freeze(self)

    @property
    def available(self) -> bool:
        try:
            return self._check_availability()

        except Exception:
            return False

    async def resolve(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Resource:
        with ctx.scope(f"resource:{self.uri}"):
            ctx.record(ArgumentsTrace.of(*args, **kwargs))
            try:
                result = Resource(
                    uri=self.uri,
                    content=await super().__call__(*args, **kwargs),  # pyright: ignore[reportCallIssue],
                )
                ctx.record(ResultTrace.of(result))

                return result

            except BaseException as exc:
                ctx.record(ResultTrace.of(exc))
                ctx.log_error(
                    "Resource resolving error",
                    exception=exc,
                )
                raise exc


def resource[**Args, Result: Sequence[Resource] | ResourceContent](
    *,
    uri: str,
    mime_type: str | None = None,
    name: str | None = None,
    description: str | None = None,
    availability_check: ResourceAvailabilityCheck | None = None,
) -> Callable[
    [Callable[Args, Coroutine[None, None, Result]]],
    ResourceTemplate[Args, Result],
]:
    def wrap(
        function: Callable[Args, Coroutine[None, None, Result]],
    ) -> ResourceTemplate[Args, Result]:
        return ResourceTemplate[Args, Result](
            uri,
            mime_type=mime_type,
            name=name or function.__name__,
            description=description,
            availability_check=availability_check,
            function=function,
        )

    return wrap



# exporting haiway symbols for easier usage
from haiway import (
    MISSING,
    AsyncQueue,
    AttributePath,
    AttributeRequirement,
    Disposable,
    Disposables,
    MetricsContext,
    MetricsHandler,
    MetricsLogger,
    MetricsRecording,
    MetricsScopeEntering,
    MetricsScopeExiting,
    Missing,
    MissingContext,
    MissingState,
    ResultTrace,
    ScopeIdentifier,
    State,
    always,
    as_dict,
    as_list,
    as_tuple,
    async_always,
    async_noop,
    asynchronous,
    cache,
    ctx,
    freeze,
    frozenlist,
    getenv_bool,
    getenv_float,
    getenv_int,
    getenv_str,
    is_missing,
    load_env,
    noop,
    not_missing,
    retry,
    setup_logging,
    throttle,
    timeout,
    traced,
    when_missing,
    wrap_async,
)

from draive.agents import (
    Agent,
    AgentError,
    AgentException,
    AgentInvocation,
    AgentMessage,
    AgentNode,
    AgentOutput,
    AgentWorkflow,
    AgentWorkflowInput,
    AgentWorkflowInvocation,
    AgentWorkflowOutput,
    agent,
    workflow,
)
from draive.choice import (
    Choice,
    ChoiceCompletion,
    ChoiceOption,
    SelectionException,
    choice_completion,
    default_choice_completion,
)
from draive.conversation import (
    Conversation,
    ConversationElement,
    ConversationMemory,
    ConversationMessage,
    conversation_completion,
    default_conversation_completion,
)
from draive.embedding import (
    Embedded,
    ImageEmbedding,
    TextEmbedding,
    ValueEmbedder,
    embed_image,
    embed_images,
    embed_text,
    embed_texts,
)
from draive.generation import (
    ImageGeneration,
    ImageGenerator,
    ModelGeneration,
    ModelGenerator,
    ModelGeneratorDecoder,
    TextGeneration,
    TextGenerator,
    generate_image,
    generate_model,
    generate_text,
)
from draive.helpers import (
    VectorIndex,
    VectorIndexing,
    VectorSearching,
    VolatileMemory,
)
from draive.instructions import (
    Instruction,
    InstructionFetching,
    InstructionsRepository,
    InstructionTemplate,
    MissingInstruction,
    fetch_instruction,
    instruction,
    instructions_file,
)
from draive.lmm import (
    LMMCompletion,
    LMMContext,
    LMMContextElement,
    LMMInput,
    LMMInvocating,
    LMMInvocation,
    LMMStream,
    LMMStreamChunk,
    LMMStreaming,
    LMMStreamInput,
    LMMStreamOutput,
    LMMStreamProperties,
    LMMToolException,
    LMMToolRequest,
    LMMToolResponse,
    LMMToolResponses,
    lmm_invoke,
    lmm_stream,
)
from draive.metrics import TokenUsage
from draive.multimodal import (
    MEDIA_KINDS,
    MEDIA_TYPES,
    MediaContent,
    MediaKind,
    MediaType,
    Multimodal,
    MultimodalContent,
    MultimodalContentConvertible,
    MultimodalContentElement,
    MultimodalTagElement,
    TextContent,
    validated_media_kind,
    validated_media_type,
)
from draive.parameters import (
    Argument,
    BasicValue,
    DataModel,
    Field,
    ParameterValidationContext,
    ParameterValidationError,
    ParameterValidator,
    ParameterVerification,
)
from draive.prompts import (
    MissingPrompt,
    Prompt,
    PromptAvailabilityCheck,
    PromptDeclaration,
    PromptDeclarationArgument,
    PromptFetching,
    PromptListing,
    PromptRepository,
    PromptTemplate,
    fetch_prompt,
    fetch_prompt_list,
    prompt,
)
from draive.resources import (
    MissingResource,
    Resource,
    ResourceContent,
    ResourceDeclaration,
    ResourceFetching,
    ResourceListing,
    ResourceRepository,
    ResourceTemplate,
    fetch_resource,
    fetch_resource_list,
    resource,
)
from draive.safeguards import (
    ContentGuardrails,
    GuardrailsException,
)
from draive.similarity import (
    mmr_vector_similarity_search,
    vector_similarity_score,
    vector_similarity_search,
)
from draive.splitters import split_text
from draive.steps import (
    Step,
    Steps,
    StepsCompleting,
    default_steps_completion,
    steps_completion,
)
from draive.tokenization import TextTokenizing, Tokenization, count_text_tokens, tokenize_text
from draive.tools import (
    ExternalToolbox,
    Tool,
    ToolAvailabilityCheck,
    Toolbox,
    ToolboxFetching,
    tool,
)
from draive.utils import (
    AsyncStream,
    ConstantStream,
    FixedStream,
    Memory,
    RateLimitError,
    split_sequence,
)

__all__ = [
    "MEDIA_KINDS",
    "MEDIA_TYPES",
    "MISSING",
    "Agent",
    "AgentError",
    "AgentException",
    "AgentInvocation",
    "AgentMessage",
    "AgentNode",
    "AgentOutput",
    "AgentWorkflow",
    "AgentWorkflowInput",
    "AgentWorkflowInvocation",
    "AgentWorkflowOutput",
    "Argument",
    "AsyncQueue",
    "AsyncStream",
    "AttributePath",
    "AttributeRequirement",
    "BasicValue",
    "Choice",
    "ChoiceCompletion",
    "ChoiceOption",
    "ConstantStream",
    "ContentGuardrails",
    "Conversation",
    "ConversationElement",
    "ConversationMemory",
    "ConversationMessage",
    "DataModel",
    "Disposable",
    "Disposables",
    "Embedded",
    "ExternalToolbox",
    "Field",
    "FixedStream",
    "GuardrailsException",
    "ImageEmbedding",
    "ImageGeneration",
    "ImageGenerator",
    "Instruction",
    "InstructionFetching",
    "InstructionTemplate",
    "InstructionsRepository",
    "LMMCompletion",
    "LMMContext",
    "LMMContextElement",
    "LMMInput",
    "LMMInvocating",
    "LMMInvocation",
    "LMMStream",
    "LMMStreamChunk",
    "LMMStreamInput",
    "LMMStreamOutput",
    "LMMStreamProperties",
    "LMMStreaming",
    "LMMToolException",
    "LMMToolException",
    "LMMToolRequest",
    "LMMToolResponse",
    "LMMToolResponses",
    "MediaContent",
    "MediaKind",
    "MediaType",
    "Memory",
    "MetricsContext",
    "MetricsHandler",
    "MetricsLogger",
    "MetricsRecording",
    "MetricsScopeEntering",
    "MetricsScopeExiting",
    "Missing",
    "MissingContext",
    "MissingInstruction",
    "MissingInstruction",
    "MissingPrompt",
    "MissingResource",
    "MissingState",
    "ModelGeneration",
    "ModelGenerator",
    "ModelGeneratorDecoder",
    "Multimodal",
    "MultimodalContent",
    "MultimodalContentConvertible",
    "MultimodalContentElement",
    "MultimodalTagElement",
    "ParameterValidationContext",
    "ParameterValidationError",
    "ParameterValidator",
    "ParameterVerification",
    "Prompt",
    "PromptAvailabilityCheck",
    "PromptDeclaration",
    "PromptDeclarationArgument",
    "PromptFetching",
    "PromptListing",
    "PromptRepository",
    "PromptTemplate",
    "RateLimitError",
    "Resource",
    "ResourceContent",
    "ResourceDeclaration",
    "ResourceFetching",
    "ResourceListing",
    "ResourceRepository",
    "ResourceTemplate",
    "ResultTrace",
    "ScopeIdentifier",
    "SelectionException",
    "State",
    "Step",
    "Steps",
    "StepsCompleting",
    "TextContent",
    "TextEmbedding",
    "TextGeneration",
    "TextGenerator",
    "TextTokenizing",
    "TokenUsage",
    "Tokenization",
    "Tool",
    "ToolAvailabilityCheck",
    "Toolbox",
    "ToolboxFetching",
    "ValueEmbedder",
    "VectorIndex",
    "VectorIndexing",
    "VectorSearching",
    "VolatileMemory",
    "agent",
    "always",
    "as_dict",
    "as_list",
    "as_tuple",
    "async_always",
    "async_noop",
    "asynchronous",
    "cache",
    "choice_completion",
    "conversation_completion",
    "count_text_tokens",
    "ctx",
    "default_choice_completion",
    "default_conversation_completion",
    "default_steps_completion",
    "embed_image",
    "embed_images",
    "embed_text",
    "embed_texts",
    "fetch_instruction",
    "fetch_prompt",
    "fetch_prompt_list",
    "fetch_resource",
    "fetch_resource_list",
    "freeze",
    "frozenlist",
    "generate_image",
    "generate_model",
    "generate_text",
    "getenv_bool",
    "getenv_float",
    "getenv_int",
    "getenv_str",
    "instruction",
    "instructions_file",
    "is_missing",
    "lmm_invoke",
    "lmm_stream",
    "load_env",
    "mmr_vector_similarity_search",
    "noop",
    "not_missing",
    "prompt",
    "resource",
    "retry",
    "setup_logging",
    "split_sequence",
    "split_text",
    "steps_completion",
    "throttle",
    "timeout",
    "tokenize_text",
    "tool",
    "traced",
    "validated_media_kind",
    "validated_media_type",
    "vector_similarity_score",
    "vector_similarity_search",
    "when_missing",
    "workflow",
    "wrap_async",
]



from typing import Any

from haiway import ctx
from openai.types.image import Image

from draive.generation import ImageGeneration
from draive.instructions import Instruction
from draive.multimodal import (
    MediaContent,
    Multimodal,
)
from draive.openai.client import OpenAIClient
from draive.openai.config import OpenAIImageGenerationConfig
from draive.openai.types import OpenAIException

__all__ = [
    "openai_image_generator",
]


def openai_image_generator(
    client: OpenAIClient | None = None,
    /,
) -> ImageGeneration:
    client = client or OpenAIClient.shared()

    async def openai_generate_image(
        *,
        instruction: Instruction | str,
        input: Multimodal | None,  # noqa: A002
        **extra: Any,
    ) -> MediaContent:
        with ctx.scope("generate_image"):
            config: OpenAIImageGenerationConfig = ctx.state(OpenAIImageGenerationConfig).updated(
                **extra
            )
            image: Image = await client.generate_image(
                config=config,
                instruction=Instruction.formatted(instruction),
            )
            if url := image.url:
                return MediaContent.url(
                    url,
                    media="image/png",
                )

            elif b64data := image.b64_json:
                return MediaContent.base64(
                    b64data,
                    media="image/png",
                )

            else:
                raise OpenAIException("Invalid OpenAI response - missing image content")

    return ImageGeneration(generate=openai_generate_image)



import json
from collections.abc import Iterable
from itertools import chain
from typing import Any

from haiway import ArgumentsTrace, ResultTrace, ctx

from draive.instructions import Instruction
from draive.lmm import (
    LMMCompletion,
    LMMContext,
    LMMContextElement,
    LMMInput,
    LMMInvocation,
    LMMOutput,
    LMMOutputSelection,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponses,
    LMMToolSelection,
    LMMToolSpecification,
)
from draive.metrics import TokenUsage
from draive.mistral.client import MistralClient
from draive.mistral.config import MistralChatConfig
from draive.mistral.models import ChatCompletionResponse, ChatMessage, ChatMessageResponse
from draive.mistral.types import MistralException
from draive.multimodal import MultimodalContent

__all__ = [
    "mistral_lmm",
]


def mistral_lmm(
    client: MistralClient | None = None,
    /,
) -> LMMInvocation:
    client = client or MistralClient.shared()

    async def lmm_invocation(  # noqa: PLR0913
        *,
        instruction: Instruction | str | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        prefill: MultimodalContent | None = None,
        **extra: Any,
    ) -> LMMOutput:
        with ctx.scope("mistral_lmm_invocation"):
            ctx.record(
                ArgumentsTrace.of(
                    instruction=instruction,
                    context=context,
                    tools=tools,
                    tool_selection=tool_selection,
                    output=output,
                    **extra,
                )
            )
            config: MistralChatConfig = ctx.state(MistralChatConfig).updated(**extra)
            ctx.record(config)

            response_format: dict[str, str]
            match output:
                case "auto" | "text":
                    response_format = {"type": "text"}

                case "image":
                    raise NotImplementedError("image output is not supported by mistral")

                case "audio":
                    raise NotImplementedError("audio output is not supported by mistral")

                case "video":
                    raise NotImplementedError("video output is not supported by mistral")

                case _:
                    if tools:
                        ctx.log_warning(
                            "Attempting to use Mistral in JSON mode with tools which is not"
                            " supported. Using text mode instead..."
                        )
                        response_format = {"type": "text"}

                    else:
                        response_format = {"type": "json_object"}


            if prefill:
                context = [*context, LMMCompletion.of(prefill)]

            messages: list[ChatMessage] = list(
                chain.from_iterable(
                    [_convert_context_element(element=element) for element in context]
                )
            )

            if instruction:
                messages = [
                    ChatMessage(
                        role="system",
                        content=Instruction.formatted(instruction),
                    ),
                    *messages,
                ]

            return await _chat_completion(
                client=client,
                config=config,
                messages=messages,
                response_format=response_format,
                tools=tools,
                tool_selection=tool_selection,
            )

    return LMMInvocation(invoke=lmm_invocation)


def _convert_context_element(
    element: LMMContextElement,
) -> Iterable[ChatMessage]:
    match element:
        case LMMInput() as input:
            return (
                ChatMessage(
                    role="user",
                    content=input.content.as_string(),
                ),
            )

        case LMMCompletion() as completion:
            return (
                ChatMessage(
                    role="assistant",
                    content=completion.content.as_string(),
                ),
            )

        case LMMToolRequests() as tool_requests:
            return (
                ChatMessage(
                    role="assistant",
                    content="",
                    tool_calls=[
                        {
                            "id": request.identifier,
                            "function": {
                                "name": request.tool,
                                "arguments": json.dumps(dict(request.arguments)),
                            },
                        }
                        for request in tool_requests.requests
                    ],
                ),
            )

        case LMMToolResponses() as tool_responses:
            return (
                ChatMessage(
                    role="tool",
                    tool_call_id=response.identifier,
                    name=response.tool,
                    content=response.content.as_string(),
                )
                for response in tool_responses.responses
            )


async def _chat_completion(  # noqa: PLR0913
    *,
    client: MistralClient,
    config: MistralChatConfig,
    messages: list[ChatMessage],
    response_format: dict[str, str],
    tools: Iterable[LMMToolSpecification] | None,
    tool_selection: LMMToolSelection,
) -> LMMOutput:
    completion: ChatCompletionResponse
    match tool_selection:
        case "auto":
            completion = await client.chat_completion(
                config=config,
                messages=messages,
                response_format=response_format,
                tools=[
                    {
                        "type": "function",
                        "function": {
                            "name": tool["name"],
                            "description": tool["description"] or "",
                            "parameters": tool["parameters"],
                        },
                    }
                    for tool in tools
                ]
                if tools
                else None,
                tool_choice="auto",
            )

        case "none":
            completion = await client.chat_completion(
                config=config,
                messages=messages,
                response_format=response_format,
                tools=[],
                tool_choice="none",
            )

        case "required":
            completion = await client.chat_completion(
                config=config,
                messages=messages,
                response_format=response_format,
                tools=[
                    {
                        "type": "function",
                        "function": {
                            "name": tool["name"],
                            "description": tool["description"] or "",
                            "parameters": tool["parameters"],
                        },
                    }
                    for tool in tools
                ]
                if tools
                else None,
                tool_choice="any",
            )

        case tool:
            assert tool in (tools or []), "Can't suggest a tool without using it"  # nosec: B101
            completion = await client.chat_completion(
                config=config,
                messages=messages,
                response_format=response_format,
                tools=[  # mistral can't be suggested with concrete tool
                    {
                        "type": "function",
                        "function": {
                            "name": tool["name"],
                            "description": tool["description"] or "",
                            "parameters": tool["parameters"],
                        },
                    }
                ],
                tool_choice="any",
            )

    if usage := completion.usage:
        ctx.record(
            TokenUsage.for_model(
                config.model,
                input_tokens=usage.prompt_tokens,
                output_tokens=usage.completion_tokens,
            ),
        )

    if not completion.choices:
        raise MistralException("Invalid Mistral completion - missing messages!", completion)

    completion_message: ChatMessageResponse = completion.choices[0].message

    if (tool_calls := completion_message.tool_calls) and (tools := tools):
        ctx.record(ResultTrace.of(tool_calls))

        return LMMToolRequests(
            requests=[
                LMMToolRequest(
                    identifier=call.id,
                    tool=call.function.name,
                    arguments=json.loads(call.function.arguments)
                    if isinstance(call.function.arguments, str)
                    else call.function.arguments,
                )
                for call in tool_calls
            ]
        )

    elif message := completion_message.content:
        ctx.record(ResultTrace.of(message))
        match message:
            case str(content):
                return LMMCompletion.of(content)

            # API docs say that it can be only a string in response
            # however library model allows list as well
            case other:
                return LMMCompletion.of(*other)

    else:
        raise MistralException("Invalid Mistral completion", completion)



from collections.abc import Mapping

from draive.parameters import DataModel

__all__ = [
    "TextContent",
]


class TextContent(DataModel):
    text: str
    meta: Mapping[str, str | float | int | bool | None] | None = None

    def __bool__(self) -> bool:
        return len(self.text) > 0



from asyncio import gather
from collections.abc import Sequence
from itertools import chain
from random import uniform
from types import TracebackType
from typing import ClassVar, Literal, Self, cast, final, overload

from haiway import getenv_str, not_missing
from openai import AsyncAzureOpenAI, AsyncOpenAI, AsyncStream
from openai import RateLimitError as OpenAIRateLimitError
from openai._types import NOT_GIVEN, NotGiven
from openai.types import Moderation, ModerationCreateResponse
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionAudioParam,
    ChatCompletionChunk,
    ChatCompletionMessageParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionToolParam,
)
from openai.types.create_embedding_response import CreateEmbeddingResponse
from openai.types.image import Image
from openai.types.images_response import ImagesResponse

from draive.openai.config import (
    OpenAIChatConfig,
    OpenAIEmbeddingConfig,
    OpenAIImageGenerationConfig,
)
from draive.parameters import DataModel
from draive.utils import RateLimitError

__all__ = [
    "OpenAIClient",
]


@final
class OpenAIClient:
    _SHARED: ClassVar[Self]

    @classmethod
    def shared(cls) -> Self:
        if shared := getattr(cls, "_SHARED", None):
            return shared

        else:
            cls._SHARED = cls()  # pyright: ignore[reportConstantRedefinition]
            return cls._SHARED

    def __init__(  # noqa: PLR0913
        self,
        base_url: str | None = None,
        api_key: str | None = None,
        organization: str | None = None,
        azure_api_endpoint: str | None = None,
        azure_api_version: str | None = None,
        azure_deployment: str | None = None,
    ) -> None:
        self._base_url: str | None = base_url or getenv_str("OPENAI_BASE_URL")
        self._api_key: str = api_key or getenv_str(
            "AZURE_OPENAI_API_KEY",
            default=getenv_str(
                "OPENAI_API_KEY",
                default="",
            ),
        )
        self._organization: str | None = organization or getenv_str("OPENAI_ORGANIZATION")
        self._azure_api_endpoint: str | None = azure_api_endpoint or getenv_str(
            "AZURE_OPENAI_API_BASE"
        )
        self._azure_deployment: str | None = azure_deployment or getenv_str(
            "AZURE_OPENAI_DEPLOYMENT_NAME"
        )
        self._azure_api_version: str | None = azure_api_version or getenv_str(
            "AZURE_OPENAI_API_VERSION"
        )
        self._client: AsyncOpenAI = self._initialize_client()

    def _initialize_client(self) -> AsyncOpenAI:
        # if AZURE settings were provided use it as provider
        if self._azure_api_endpoint:
            return AsyncAzureOpenAI(
                api_key=self._api_key,
                azure_endpoint=self._azure_api_endpoint,
                azure_deployment=self._azure_deployment,
                api_version=self._azure_api_version,
                organization=self._organization,
            )

        else:  # otherwise try using OpenAI
            return AsyncOpenAI(
                base_url=self._base_url,
                api_key=self._api_key,
                organization=self._organization,
            )

    @overload
    async def chat_completion(
        self,
        *,
        config: OpenAIChatConfig,
        messages: list[ChatCompletionMessageParam],
        tools: list[ChatCompletionToolParam] | None = None,
        tool_choice: ChatCompletionToolChoiceOptionParam | NotGiven = NOT_GIVEN,
        stream: Literal[True],
    ) -> AsyncStream[ChatCompletionChunk]: ...

    @overload
    async def chat_completion(
        self,
        *,
        config: OpenAIChatConfig,
        messages: list[ChatCompletionMessageParam],
        response_format: Literal["auto", "text", "image", "audio", "video"]
        | type[DataModel] = "auto",
        tools: list[ChatCompletionToolParam] | None = None,
        tool_choice: ChatCompletionToolChoiceOptionParam | NotGiven = NOT_GIVEN,
    ) -> ChatCompletion: ...

    async def chat_completion(  # noqa: PLR0913
        self,
        *,
        config: OpenAIChatConfig,
        messages: list[ChatCompletionMessageParam],
        response_format: Literal["auto", "text", "image", "audio", "video"]
        | type[DataModel] = "auto",
        tools: list[ChatCompletionToolParam] | None = None,
        tool_choice: ChatCompletionToolChoiceOptionParam | NotGiven = NOT_GIVEN,
        stream: bool = False,
    ) -> AsyncStream[ChatCompletionChunk] | ChatCompletion:
        try:
            return await self._client.chat.completions.create(
                messages=messages,
                model=config.model,
                modalities=["audio"] if response_format == "audio" else NOT_GIVEN,
                audio=ChatCompletionAudioParam(**config.audio_response_format)
                if not_missing(config.audio_response_format)
                else NOT_GIVEN,
                frequency_penalty=config.frequency_penalty
                if not_missing(config.frequency_penalty)
                else NOT_GIVEN,
                max_tokens=config.max_tokens if not_missing(config.max_tokens) else NOT_GIVEN,
                n=1,
                response_format={
                    # TODO: allow json schema: response_format.__PARAMETERS_SPECIFICATION__
                    "type": "json_object",
                }
                if isinstance(response_format, type)
                else NOT_GIVEN,
                seed=config.seed if not_missing(config.seed) else NOT_GIVEN,
                stream=stream,
                temperature=config.temperature,
                tools=tools or NOT_GIVEN,
                tool_choice=tool_choice if tools else NOT_GIVEN,
                parallel_tool_calls=(
                    config.parallel_tool_calls
                    if not_missing(config.parallel_tool_calls)
                    else NOT_GIVEN
                )
                if tools
                else NOT_GIVEN,
                top_p=config.top_p if not_missing(config.top_p) else NOT_GIVEN,
                timeout=config.timeout if not_missing(config.timeout) else NOT_GIVEN,
                stream_options={"include_usage": True} if stream else NOT_GIVEN,
                stop=list(config.stop_sequences)
                if not_missing(config.stop_sequences)
                else NOT_GIVEN,
            )

        except OpenAIRateLimitError as exc:  # retry on rate limit after delay
            if delay := exc.response.headers.get("Retry-After"):
                try:
                    raise RateLimitError(
                        retry_after=float(delay) + uniform(0.0, 0.3)  # nosec: B311 # add small random delay
                    ) from exc

                except ValueError:
                    raise exc from None

            else:
                raise exc

    async def embedding(
        self,
        config: OpenAIEmbeddingConfig,
        inputs: Sequence[str],
    ) -> list[list[float]]:
        return list(
            chain(
                *await gather(
                    *[
                        self._create_text_embedding(
                            texts=inputs[index : index + config.batch_size],
                            model=config.model,
                            dimensions=config.dimensions
                            if not_missing(config.dimensions)
                            else NOT_GIVEN,
                            encoding_format=cast(Literal["float", "base64"], config.encoding_format)
                            if not_missing(config.encoding_format)
                            else NOT_GIVEN,
                            timeout=config.timeout if not_missing(config.timeout) else NOT_GIVEN,
                        )
                        for index in range(0, len(inputs), config.batch_size)
                    ]
                )
            )
        )

    async def _create_text_embedding(
        self,
        texts: Sequence[str],
        model: str,
        dimensions: int | NotGiven,
        encoding_format: Literal["float", "base64"] | NotGiven,
        timeout: float | NotGiven,
    ) -> list[list[float]]:
        while True:
            try:
                response: CreateEmbeddingResponse = await self._client.embeddings.create(
                    input=list(texts),
                    model=model,
                    dimensions=dimensions,
                    encoding_format=encoding_format,
                    timeout=timeout,
                )
                return [element.embedding for element in response.data]

            except OpenAIRateLimitError as exc:  # always retry on rate limit after delay
                if delay := exc.response.headers.get("Retry-After"):
                    try:
                        raise RateLimitError(
                            retry_after=float(delay) + uniform(0.0, 0.3)  # nosec: B311 # add small random delay
                        ) from exc

                    except ValueError:
                        raise exc from None

                else:
                    raise exc

    async def moderation_check(
        self,
        text: str,
    ) -> Moderation:
        response: ModerationCreateResponse = await self._client.moderations.create(
            input=text,
        )
        return response.results[0]  # TODO: check API about multiple results

    async def generate_image(
        self,
        config: OpenAIImageGenerationConfig,
        instruction: str,
    ) -> Image:
        response: ImagesResponse = await self._client.images.generate(
            model=config.model,
            n=1,
            prompt=instruction,
            quality=config.quality,
            size=config.size,
            style=config.style,
            timeout=config.timeout if not_missing(config.timeout) else NOT_GIVEN,
            response_format=config.response_format,
        )
        return response.data[0]

    async def __aenter__(self) -> None:
        if self._client.is_closed():
            self._client = self._initialize_client()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._client.close()



from base64 import b64encode
from collections.abc import Iterable
from typing import Any, cast

from anthropic.types import (
    ImageBlockParam,
    Message,
    MessageParam,
    TextBlock,
    TextBlockParam,
    ToolParam,
    ToolUseBlock,
)
from haiway import ArgumentsTrace, ResultTrace, ctx

from draive.anthropic.client import AnthropicClient
from draive.anthropic.config import AnthropicConfig
from draive.anthropic.types import AnthropicException
from draive.instructions import Instruction
from draive.lmm import (
    LMMCompletion,
    LMMContext,
    LMMContextElement,
    LMMInput,
    LMMInvocation,
    LMMOutput,
    LMMOutputSelection,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponses,
    LMMToolSelection,
    LMMToolSpecification,
)
from draive.metrics import TokenUsage
from draive.multimodal import MediaContent, MultimodalContent, MultimodalContentElement, TextContent
from draive.parameters import DataModel

__all__ = [
    "anthropic_lmm",
]


def anthropic_lmm(
    client: AnthropicClient | None = None,
    /,
) -> LMMInvocation:
    client = client or AnthropicClient.shared()

    async def lmm_invocation(  # noqa: PLR0913
        *,
        instruction: Instruction | str | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        prefill: MultimodalContent | None = None,
        **extra: Any,
    ) -> LMMOutput:
        with ctx.scope("anthropic_lmm_invocation"):
            ctx.record(
                ArgumentsTrace.of(
                    instruction=instruction,
                    context=context,
                    prefill=prefill,
                    tool_selection=tool_selection,
                    tools=tools,
                    output=output,
                    **extra,
                )
            )
            config: AnthropicConfig = ctx.state(AnthropicConfig).updated(**extra)
            ctx.record(config)

            match output:
                case "auto" | "text":
                    pass

                case "image":
                    raise NotImplementedError("image output is not supported by anthropic")

                case "audio":
                    raise NotImplementedError("audio output is not supported by anthropic")

                case "video":
                    raise NotImplementedError("video output is not supported by anthropic")

                case _:
                    pass  # model output is not normalized but we can use prefill

            if prefill:
                context = [*context, LMMCompletion.of(prefill)]

            return await _completion(
                client=client,
                config=config,
                instruction=Instruction.formatted(instruction),
                messages=[_convert_context_element(element=element) for element in context],
                tools=tools,
                tool_selection=tool_selection,
            )

    return LMMInvocation(invoke=lmm_invocation)


def _convert_content_element(
    element: MultimodalContentElement,
) -> TextBlockParam | ImageBlockParam:
    match element:
        case TextContent() as text:
            return {
                "type": "text",
                "text": text.text,
            }

        case MediaContent() as media:
            if media.kind != "image" or isinstance(media.source, str):
                raise ValueError("Unsupported message content", media)

            return {
                "type": "image",
                "source": {
                    "type": "base64",
                    "media_type": cast(Any, media.media),
                    "data": b64encode(media.source).decode(),
                },
            }

        case DataModel() as data:
            return {
                "type": "text",
                "text": data.as_json(),
            }


def _convert_context_element(
    element: LMMContextElement,
) -> MessageParam:
    match element:
        case LMMInput() as input:
            return {
                "role": "user",
                "content": [
                    _convert_content_element(element=element) for element in input.content.parts
                ],
            }

        case LMMCompletion() as completion:
            # TODO: Anthropic models generating media?
            return {
                "role": "assistant",
                "content": completion.content.as_string(),
            }

        case LMMToolRequests() as tool_requests:
            return {
                "role": "assistant",
                "content": [
                    {
                        "id": request.identifier,
                        "type": "tool_use",
                        "name": request.tool,
                        "input": request.arguments,
                    }
                    for request in tool_requests.requests
                ],
            }

        case LMMToolResponses() as tool_responses:
            return {
                "role": "user",
                "content": [
                    {
                        "tool_use_id": response.identifier,
                        "type": "tool_result",
                        "is_error": response.error,
                        "content": [
                            _convert_content_element(element=part)
                            for part in response.content.parts
                        ],
                    }
                    for response in tool_responses.responses
                ],
            }


async def _completion(  # noqa: PLR0913, PLR0912, C901
    *,
    client: AnthropicClient,
    config: AnthropicConfig,
    instruction: str | None,
    messages: list[MessageParam],
    tools: Iterable[LMMToolSpecification] | None,
    tool_selection: LMMToolSelection,
) -> LMMOutput:
    completion: Message
    match tool_selection:
        case "auto":
            completion = await client.completion(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[
                    ToolParam(
                        name=tool["name"],
                        description=tool["description"] or "",
                        input_schema=cast(
                            dict[str, Any],
                            tool["parameters"],
                        ),
                    )
                    for tool in tools or []
                ],
                tool_choice="auto",
            )

        case "none":
            completion = await client.completion(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[
                    ToolParam(
                        name=tool["name"],
                        description=tool["description"] or "",
                        input_schema=cast(
                            dict[str, Any],
                            tool["parameters"],
                        ),
                    )
                    for tool in tools or []
                ],
                tool_choice="none",
            )

        case "required":
            completion = await client.completion(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[
                    ToolParam(
                        name=tool["name"],
                        description=tool["description"] or "",
                        input_schema=cast(
                            dict[str, Any],
                            tool["parameters"],
                        ),
                    )
                    for tool in tools or []
                ],
                tool_choice="any",
            )

        case tool:
            completion = await client.completion(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[
                    ToolParam(
                        name=tool["name"],
                        description=tool["description"] or "",
                        input_schema=cast(
                            dict[str, Any],
                            tool["parameters"],
                        ),
                    )
                    for tool in tools or []
                ],
                tool_choice={
                    "type": "tool",
                    "name": tool["name"],
                },
            )

    ctx.record(
        TokenUsage.for_model(
            config.model,
            input_tokens=completion.usage.input_tokens or 0,
            output_tokens=completion.usage.output_tokens or 0,
        ),
    )

    message_parts: list[TextBlock]
    match messages[-1]:
        case {"role": "assistant", "content": str() as content_text}:
            message_parts = [TextBlock(type="text", text=content_text)]

        case {"role": "assistant", "content": content_parts}:
            message_parts = [  # currently supporting only text prefills
                TextBlock(type="text", text=part.text)
                for part in content_parts
                if isinstance(part, TextBlock)
            ]

        case _:
            message_parts = []

    tool_calls: list[ToolUseBlock] = []
    for part in completion.content:
        match part:
            case TextBlock() as text:
                message_parts.append(text)

            case ToolUseBlock() as call:
                tool_calls.append(call)

    match completion.stop_reason:
        case "tool_use":
            if (tool_calls := tool_calls) and (tools := tools):
                ctx.record(ResultTrace.of(tool_calls))
                return LMMToolRequests(
                    requests=[
                        LMMToolRequest(
                            identifier=call.id,
                            tool=call.name,
                            arguments=cast(dict[str, Any], call.input),
                        )
                        for call in tool_calls
                    ]
                )

            else:
                raise AnthropicException("Invalid Anthropic completion", completion)

        case "end_turn" | "stop_sequence":
            if (tool_calls := tool_calls) and (tools := tools):
                ctx.record(ResultTrace.of(tool_calls))
                return LMMToolRequests(
                    requests=[
                        LMMToolRequest(
                            identifier=call.id,
                            tool=call.name,
                            arguments=cast(dict[str, Any], call.input),
                        )
                        for call in tool_calls
                    ]
                )

            else:
                ctx.record(ResultTrace.of(message_parts))

                return LMMCompletion.of(
                    MultimodalContent.of(
                        *[TextContent(text=part.text) for part in message_parts],
                    )
                )

        case other:
            raise AnthropicException(f"Unexpected finish reason: {other}")



from collections.abc import Sequence
from types import TracebackType
from typing import Any, ClassVar, Self, final

from haiway import asynchronous, getenv_str, not_missing

from draive.bedrock.config import BedrockChatConfig
from draive.bedrock.models import ChatCompletionResponse, ChatMessage, ChatTool

__all__ = [
    "BedrockClient",
]


@final
class BedrockClient:
    _SHARED: ClassVar[Self]

    @classmethod
    def shared(cls) -> Self:
        if shared := getattr(cls, "_SHARED", None):
            return shared

        else:
            cls._SHARED = cls()  # pyright: ignore[reportConstantRedefinition]
            return cls._SHARED

    def __init__(
        self,
        region_name: str | None = None,
        access_key_id: str | None = None,
        access_key: str | None = None,
    ) -> None:
        self._region_name: str | None = region_name or getenv_str("AWS_DEFAULT_REGION")
        self._access_key_id: str | None = access_key_id or getenv_str("AWS_ACCESS_KEY_ID")
        self._access_key: str | None = access_key or getenv_str("AWS_ACCESS_KEY")
        self._client: Any

    # preparing it lazily on demand, boto does a lot of stuff on initialization
    @asynchronous
    def initialize(self) -> None:
        if hasattr(self, "_client"):
            return  # already initialized

        import boto3  # pyright: ignore[reportMissingTypeStubs]

        self._client = boto3.Session(  # pyright: ignore[reportUnknownMemberType]
            aws_access_key_id=self._access_key_id,
            aws_secret_access_key=self._access_key,
            region_name=self._region_name,
        ).client("bedrock-runtime")

    async def chat_completion(
        self,
        *,
        config: BedrockChatConfig,
        instruction: str | None,
        messages: list[ChatMessage],
        tools: list[ChatTool],
        require_tool: str | bool,
    ) -> ChatCompletionResponse:
        return await self._create_chat_completion(
            model=config.model,
            temperature=config.temperature,
            top_p=config.top_p if not_missing(config.top_p) else None,
            max_tokens=config.max_tokens if not_missing(config.max_tokens) else None,
            instruction=instruction,
            messages=messages,
            tools=tools,
            require_tool=require_tool,
            stop=config.stop_sequences if not_missing(config.stop_sequences) else None,
        )

    @asynchronous
    def _create_chat_completion(  # noqa: PLR0913
        self,
        model: str,
        temperature: float,
        top_p: float | None,
        max_tokens: int | None,
        instruction: str | None,
        messages: list[ChatMessage],
        tools: list[ChatTool],
        require_tool: str | bool,
        stop: Sequence[str] | None,
    ) -> ChatCompletionResponse:
        parameters: dict[str, Any] = {
            "modelId": model,
            "messages": messages,
            "inferenceConfig": {
                "temperature": temperature,
            },
        }

        if instruction:
            parameters["system"] = [
                {
                    "text": instruction,
                }
            ]

        if tools:
            toolChoice: Any
            match require_tool:
                case False:
                    toolChoice = {"auto": {}}

                case True:
                    toolChoice = {"any": {}}

                case tool:
                    toolChoice = {"tool": {"name": tool}}

            parameters["toolConfig"] = {
                "tools": [{"toolSpec": tool} for tool in tools],
                "toolChoice": toolChoice,
            }

        if max_tokens:
            parameters["inferenceConfig"]["maxTokens"] = max_tokens
        if top_p is not None:
            parameters["inferenceConfig"]["topP"] = top_p
        if stop:
            parameters["inferenceConfig"]["stopSequences"] = stop

        return self._client.converse(**parameters)

    async def __aenter__(self) -> None:
        await self.initialize()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        pass
